"""Groq language model provider."""

import json
import os
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Union,
)

import httpx

from esperanto.common_types import (
    ChatCompletion,
    ChatCompletionChunk,
    Choice,
    DeltaMessage,
    FunctionCall,
    Message,
    Model,
    StreamChoice,
    Tool,
    ToolCall,
    Usage,
)
from esperanto.common_types.validation import (
    validate_tool_calls as _validate_tool_calls,
)
from esperanto.providers.llm.base import LanguageModel

if TYPE_CHECKING:
    from langchain_groq import ChatGroq


class GroqLanguageModel(LanguageModel):
    """Groq language model implementation."""

    def __post_init__(self):
        """Initialize HTTP clients."""
        # Call parent's post_init to handle config initialization
        super().__post_init__()

        # Get API key
        self.api_key = self.api_key or os.getenv("GROQ_API_KEY")
        if not self.api_key:
            raise ValueError("Groq API key not found")

        # Set base URL for Groq's OpenAI-compatible API
        self.base_url = "https://api.groq.com/openai/v1"

        # Initialize HTTP clients with configurable timeout
        self._create_http_clients()

    def _get_headers(self) -> Dict[str, str]:
        """Get headers for Groq API requests."""
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _handle_error(self, response: httpx.Response) -> None:
        """Handle HTTP error responses."""
        if response.status_code >= 400:
            try:
                error_data = response.json()
                error_message = error_data.get("error", {}).get("message", f"HTTP {response.status_code}")
            except Exception:
                error_message = f"HTTP {response.status_code}: {response.text}"
            raise RuntimeError(f"Groq API error: {error_message}")

    def _get_models(self) -> List[Model]:
        """List all available models for this provider."""
        response = self.client.get(
            f"{self.base_url}/models",
            headers=self._get_headers()
        )
        self._handle_error(response)
        
        models_data = response.json()
        return [
            Model(
                id=model["id"],
                owned_by="Groq",
                context_window=128000,  # All Groq models currently support 128k context
            )
            for model in models_data["data"]
        ]

    def _convert_tools_to_openai(
        self, tools: Optional[List[Tool]]
    ) -> Optional[List[Dict[str, Any]]]:
        """Convert Esperanto tools to OpenAI/Groq format.

        Args:
            tools: List of Esperanto Tool objects.

        Returns:
            List of tools in OpenAI/Groq API format, or None if no tools provided.
        """
        if not tools:
            return None
        result = []
        for tool in tools:
            tool_dict: Dict[str, Any] = {
                "type": tool.type,
                "function": {
                    "name": tool.function.name,
                    "description": tool.function.description,
                    "parameters": tool.function.parameters,
                },
            }
            # Add strict mode if specified (OpenAI-specific, but include for compat)
            if tool.function.strict is not None:
                tool_dict["function"]["strict"] = tool.function.strict
            result.append(tool_dict)
        return result

    def _normalize_response(self, response_data: Dict[str, Any]) -> ChatCompletion:
        """Normalize Groq response to our format."""
        choices = []
        for choice in response_data["choices"]:
            message_data = choice.get("message", {})

            # Extract tool_calls if present
            tool_calls = None
            if "tool_calls" in message_data and message_data["tool_calls"]:
                tool_calls = [
                    ToolCall(
                        id=tc["id"],
                        type=tc.get("type", "function"),
                        function=FunctionCall(
                            name=tc["function"]["name"],
                            arguments=tc["function"]["arguments"],
                        ),
                    )
                    for tc in message_data["tool_calls"]
                ]

            choices.append(
                Choice(
                    index=choice["index"],
                    message=Message(
                        content=message_data.get("content") or "",
                        role=message_data.get("role", "assistant"),
                        tool_calls=tool_calls,
                    ),
                    finish_reason=choice.get("finish_reason"),
                )
            )

        return ChatCompletion(
            id=response_data["id"],
            choices=choices,
            created=response_data["created"],
            model=response_data["model"],
            provider=self.provider,
            usage=Usage(
                completion_tokens=response_data.get("usage", {}).get("completion_tokens", 0),
                prompt_tokens=response_data.get("usage", {}).get("prompt_tokens", 0),
                total_tokens=response_data.get("usage", {}).get("total_tokens", 0),
            ),
        )

    def _normalize_chunk(self, chunk_data: Dict[str, Any]) -> ChatCompletionChunk:
        """Normalize Groq stream chunk to our format."""
        return ChatCompletionChunk(
            id=chunk_data["id"],
            choices=[
                StreamChoice(
                    index=choice["index"],
                    delta=DeltaMessage(
                        content=choice.get("delta", {}).get("content", ""),
                        role=choice.get("delta", {}).get("role", "assistant"),
                        function_call=choice.get("delta", {}).get("function_call"),
                        tool_calls=choice.get("delta", {}).get("tool_calls"),
                    ),
                    finish_reason=choice.get("finish_reason"),
                )
                for choice in chunk_data["choices"]
            ],
            created=chunk_data["created"],
            model=chunk_data.get("model", ""),
        )

    def _parse_sse_stream(self, response: httpx.Response) -> Generator[Dict[str, Any], None, None]:
        """Parse Server-Sent Events stream from Groq chat completions."""
        for chunk in response.iter_text():
            for line in chunk.split('\n'):
                line = line.strip()
                if not line:
                    continue
                if line.startswith("data: "):
                    data = line[6:]  # Remove "data: " prefix
                    if data.strip() == "[DONE]":
                        break
                    try:
                        yield json.loads(data)
                    except json.JSONDecodeError:
                        continue

    async def _parse_sse_stream_async(self, response: httpx.Response) -> AsyncGenerator[Dict[str, Any], None]:
        """Parse Server-Sent Events stream from Groq chat completions asynchronously."""
        async for chunk in response.aiter_text():
            for line in chunk.split('\n'):
                line = line.strip()
                if not line:
                    continue
                if line.startswith("data: "):
                    data = line[6:]  # Remove "data: " prefix
                    if data.strip() == "[DONE]":
                        return
                    try:
                        yield json.loads(data)
                    except json.JSONDecodeError:
                        continue

    def _get_api_kwargs(self, exclude_stream: bool = False) -> Dict[str, Any]:
        """Get kwargs for API calls, filtering out provider-specific args."""
        kwargs = {}
        config = self.get_completion_kwargs()

        # Only include non-provider-specific args that were explicitly set
        for key, value in config.items():
            if key not in [
                "model_name",
                "api_key",
                "base_url",
                "organization",
                "structured",
                # Tool-related fields are handled separately in chat_complete()
                "tools",
                "tool_choice",
                "parallel_tool_calls",
            ]:
                # Skip max_tokens if it's the default value (850)
                if key == "max_tokens" and value == 850:
                    continue
                kwargs[key] = value

        # Handle streaming parameter
        if exclude_stream:
            kwargs.pop("streaming", None)
        elif "streaming" in kwargs:
            kwargs["stream"] = kwargs.pop("streaming")

        # Handle structured output
        if self.structured:
            if not isinstance(self.structured, dict):
                raise TypeError("structured parameter must be a dictionary")
            structured_type = self.structured.get("type")
            if structured_type in ["json", "json_object"]:
                kwargs["response_format"] = {"type": "json_object"}

        return kwargs

    def chat_complete(
        self,
        messages: List[Dict[str, Any]],
        stream: Optional[bool] = None,
        tools: Optional[List[Tool]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        parallel_tool_calls: Optional[bool] = None,
        validate_tool_calls: bool = False,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """Send a chat completion request.

        Args:
            messages: List of messages in the conversation. Messages can include
                tool call results with role="tool" and tool_call_id.
            stream: Whether to stream the response. If None, uses the instance's
                streaming setting.
            tools: List of tools the model can call. If None, uses instance tools.
            tool_choice: Controls tool usage. Values:
                - "auto": Model decides whether to call tools (default)
                - "required": Model must call at least one tool
                - "none": Model cannot call tools
                - {"type": "function", "function": {"name": "..."}}: Force specific tool
            parallel_tool_calls: Whether to allow multiple tool calls in one response.
                None uses provider default (usually True). Set False to force single
                tool call per response.
            validate_tool_calls: If True, validate tool call arguments against the
                tool's JSON schema. Raises ToolCallValidationError on validation
                failure. Requires jsonschema package.

        Returns:
            Either a ChatCompletion or a Generator yielding ChatCompletionChunks
            if streaming. When the model calls tools, the response message will
            have tool_calls populated.
        """
        # Warn if validate_tool_calls is used with streaming
        self._warn_if_validate_with_streaming(validate_tool_calls, stream)

        should_stream = stream if stream is not None else self.streaming

        # Resolve tool configuration
        resolved_tools = self._resolve_tools(tools)
        resolved_tool_choice = self._resolve_tool_choice(tool_choice)
        resolved_parallel = self._resolve_parallel_tool_calls(parallel_tool_calls)

        # Prepare request payload
        payload: Dict[str, Any] = {
            "model": self.get_model_name(),
            "messages": messages,
            "stream": should_stream,
            **self._get_api_kwargs(exclude_stream=True),
        }

        # Add tool-related parameters if configured
        if resolved_tools:
            payload["tools"] = self._convert_tools_to_openai(resolved_tools)
        if resolved_tool_choice is not None:
            payload["tool_choice"] = resolved_tool_choice
        if resolved_parallel is not None:
            payload["parallel_tool_calls"] = resolved_parallel

        # Make HTTP request
        response = self.client.post(
            f"{self.base_url}/chat/completions",
            headers=self._get_headers(),
            json=payload
        )
        self._handle_error(response)

        if should_stream:
            return (self._normalize_chunk(chunk_data) for chunk_data in self._parse_sse_stream(response))

        response_data = response.json()
        result = self._normalize_response(response_data)

        # Validate tool calls if requested
        if validate_tool_calls and resolved_tools:
            for choice in result.choices:
                if choice.message.tool_calls:
                    _validate_tool_calls(choice.message.tool_calls, resolved_tools)

        return result

    async def achat_complete(
        self,
        messages: List[Dict[str, Any]],
        stream: Optional[bool] = None,
        tools: Optional[List[Tool]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        parallel_tool_calls: Optional[bool] = None,
        validate_tool_calls: bool = False,
    ) -> Union[ChatCompletion, AsyncGenerator[ChatCompletionChunk, None]]:
        """Send an async chat completion request.

        Args:
            messages: List of messages in the conversation. Messages can include
                tool call results with role="tool" and tool_call_id.
            stream: Whether to stream the response. If None, uses the instance's
                streaming setting.
            tools: List of tools the model can call. If None, uses instance tools.
            tool_choice: Controls tool usage. Values:
                - "auto": Model decides whether to call tools (default)
                - "required": Model must call at least one tool
                - "none": Model cannot call tools
                - {"type": "function", "function": {"name": "..."}}: Force specific tool
            parallel_tool_calls: Whether to allow multiple tool calls in one response.
                None uses provider default (usually True). Set False to force single
                tool call per response.
            validate_tool_calls: If True, validate tool call arguments against the
                tool's JSON schema. Raises ToolCallValidationError on validation
                failure. Requires jsonschema package.

        Returns:
            Either a ChatCompletion or an AsyncGenerator yielding ChatCompletionChunks
            if streaming. When the model calls tools, the response message will
            have tool_calls populated.
        """
        # Warn if validate_tool_calls is used with streaming
        self._warn_if_validate_with_streaming(validate_tool_calls, stream)

        should_stream = stream if stream is not None else self.streaming

        # Resolve tool configuration
        resolved_tools = self._resolve_tools(tools)
        resolved_tool_choice = self._resolve_tool_choice(tool_choice)
        resolved_parallel = self._resolve_parallel_tool_calls(parallel_tool_calls)

        # Prepare request payload
        payload: Dict[str, Any] = {
            "model": self.get_model_name(),
            "messages": messages,
            "stream": should_stream,
            **self._get_api_kwargs(exclude_stream=True),
        }

        # Add tool-related parameters if configured
        if resolved_tools:
            payload["tools"] = self._convert_tools_to_openai(resolved_tools)
        if resolved_tool_choice is not None:
            payload["tool_choice"] = resolved_tool_choice
        if resolved_parallel is not None:
            payload["parallel_tool_calls"] = resolved_parallel

        # Make async HTTP request
        response = await self.async_client.post(
            f"{self.base_url}/chat/completions",
            headers=self._get_headers(),
            json=payload
        )
        self._handle_error(response)

        if should_stream:
            async def generate():
                async for chunk_data in self._parse_sse_stream_async(response):
                    yield self._normalize_chunk(chunk_data)

            return generate()

        response_data = response.json()
        result = self._normalize_response(response_data)

        # Validate tool calls if requested
        if validate_tool_calls and resolved_tools:
            for choice in result.choices:
                if choice.message.tool_calls:
                    _validate_tool_calls(choice.message.tool_calls, resolved_tools)

        return result

    def _get_default_model(self) -> str:
        """Get the default model name."""
        return "mixtral-8x7b-32768"

    @property
    def provider(self) -> str:
        """Get the provider name."""
        return "groq"

    def to_langchain(self) -> "ChatGroq":
        """Convert to a LangChain chat model.

        Raises:
            ImportError: If langchain_groq is not installed.
        """
        try:
            from langchain_groq import ChatGroq
        except ImportError as e:
            raise ImportError(
                "Langchain integration requires langchain_groq. "
                "Install with: uv add langchain_groq or pip install langchain_groq"
            ) from e

        # SecretStr import removed, rely on ChatGroq internal handling

        # Ensure model name is a string
        model_name = self.get_model_name()
        if not model_name:
            raise ValueError("Model name must be set to use Langchain integration.")

        langchain_kwargs = {
            "model": model_name,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            # top_p=self.top_p, # Still not supported
            "streaming": self.streaming,
            "api_key": self.api_key,  # Pass the raw API key string
        }

        # Create new HTTP clients for LangChain with same SSL/timeout/proxy config
        # We create fresh clients instead of sharing ours because when this Esperanto
        # model is garbage collected, __del__ closes our clients - which would break
        # LangChain if it was sharing them. Fresh clients give LangChain ownership.
        try:
            sync_client, async_client = self._create_langchain_http_clients()
            langchain_kwargs["http_client"] = sync_client
            langchain_kwargs["http_async_client"] = async_client
        except (TypeError, AttributeError):
            # httpx types might be mocked in tests, skip passing clients
            pass

        return ChatGroq(**self._clean_config(langchain_kwargs))



#!/usr/bin/env python3
"""Mistral tool calling example.

Demonstrates tool calling with Mistral AI models.
Mistral uses OpenAI-compatible format for tools.

Environment variables required:
- MISTRAL_API_KEY: Your Mistral API key

Run with:
    python examples/tool_calling/provider_examples/mistral_tools.py
"""

import json
import os
from typing import Any, Dict

from esperanto import AIFactory
from esperanto.common_types import Tool, ToolFunction


def create_tools():
    """Create tools for the example."""
    return [
        Tool(
            type="function",
            function=ToolFunction(
                name="get_flight_info",
                description="Get information about a flight",
                parameters={
                    "type": "object",
                    "properties": {
                        "flight_number": {
                            "type": "string",
                            "description": "The flight number, e.g., 'UA123'",
                        },
                        "date": {
                            "type": "string",
                            "description": "The flight date in YYYY-MM-DD format",
                        },
                    },
                    "required": ["flight_number"],
                },
            ),
        ),
        Tool(
            type="function",
            function=ToolFunction(
                name="book_hotel",
                description="Book a hotel room",
                parameters={
                    "type": "object",
                    "properties": {
                        "city": {
                            "type": "string",
                            "description": "The city for the hotel",
                        },
                        "check_in": {
                            "type": "string",
                            "description": "Check-in date (YYYY-MM-DD)",
                        },
                        "check_out": {
                            "type": "string",
                            "description": "Check-out date (YYYY-MM-DD)",
                        },
                        "guests": {
                            "type": "integer",
                            "description": "Number of guests",
                            "default": 1,
                        },
                    },
                    "required": ["city", "check_in", "check_out"],
                },
            ),
        ),
    ]


def execute_tool(name: str, args: Dict[str, Any]) -> Dict[str, Any]:
    """Simulate tool execution."""
    if name == "get_flight_info":
        return {
            "flight_number": args.get("flight_number"),
            "date": args.get("date", "2024-01-20"),
            "departure": "SFO",
            "arrival": "JFK",
            "departure_time": "08:00",
            "arrival_time": "16:30",
            "status": "On Time",
        }
    elif name == "book_hotel":
        return {
            "confirmation": "HTL-" + str(hash(args.get("city", "")))[1:7],
            "hotel": f"Grand Hotel {args.get('city')}",
            "city": args.get("city"),
            "check_in": args.get("check_in"),
            "check_out": args.get("check_out"),
            "guests": args.get("guests", 1),
            "price_per_night": 150.00,
        }
    return {"error": f"Unknown tool: {name}"}


def main():
    # Check for API key
    if not os.getenv("MISTRAL_API_KEY"):
        print("Please set MISTRAL_API_KEY environment variable")
        return

    print("=== Mistral Tool Calling Example ===\n")

    tools = create_tools()

    # Create model
    model = AIFactory.create_language(
        provider="mistral",
        model_name="mistral-small-latest",
    )

    # Example 1: Basic tool call
    print("1. Basic Tool Call")
    print("-" * 40)
    messages = [{"role": "user", "content": "What's the status of flight UA456?"}]
    response = model.chat_complete(messages, tools=tools)

    if response.choices[0].message.tool_calls:
        tc = response.choices[0].message.tool_calls[0]
        print(f"Tool: {tc.function.name}")
        print(f"Args: {json.loads(tc.function.arguments)}")
    print()

    # Example 2: Multi-turn travel booking
    print("2. Multi-turn Travel Booking")
    print("-" * 40)

    messages = [
        {
            "role": "user",
            "content": "I need to check flight AA100 for tomorrow and book a hotel in New York from Jan 20 to Jan 23",
        }
    ]

    response = model.chat_complete(messages, tools=tools)

    # Process tool calls
    iteration = 0
    while response.choices[0].message.tool_calls and iteration < 5:
        iteration += 1
        tool_calls = response.choices[0].message.tool_calls

        # Add assistant message
        messages.append(
            {
                "role": "assistant",
                "tool_calls": [
                    {
                        "id": tc.id,
                        "type": tc.type,
                        "function": {
                            "name": tc.function.name,
                            "arguments": tc.function.arguments,
                        },
                    }
                    for tc in tool_calls
                ],
            }
        )

        # Execute each tool
        for tc in tool_calls:
            args = json.loads(tc.function.arguments)
            result = execute_tool(tc.function.name, args)
            print(f"  [{tc.function.name}] {args} -> {json.dumps(result)[:60]}...")

            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tc.id,
                    "content": json.dumps(result),
                }
            )

        response = model.chat_complete(messages, tools=tools)

    print(f"\nMistral's summary: {response.choices[0].message.content}")
    print()

    # Example 3: tool_choice options
    print("3. Forcing Specific Tool")
    print("-" * 40)
    messages = [{"role": "user", "content": "I want to stay somewhere nice"}]
    response = model.chat_complete(
        messages,
        tools=tools,
        tool_choice={"type": "function", "function": {"name": "book_hotel"}},
    )

    if response.choices[0].message.tool_calls:
        tc = response.choices[0].message.tool_calls[0]
        print(f"Forced to call: {tc.function.name}")
        print(f"Args: {json.loads(tc.function.arguments)}")
    print()

    print("=== Example Complete ===")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Vertex AI tool calling example.

Demonstrates tool calling with Google Cloud Vertex AI (Gemini models).
Vertex AI uses Google's function calling format.

Requirements:
- Google Cloud project with Vertex AI enabled
- Authenticated via gcloud CLI or service account

Environment variables:
- GOOGLE_CLOUD_PROJECT: Your Google Cloud project ID
- GOOGLE_CLOUD_LOCATION: The region (default: us-central1)

Run with:
    python examples/tool_calling/provider_examples/vertex_tools.py
"""

import json
import os
from typing import Any, Dict

from esperanto import AIFactory
from esperanto.common_types import Tool, ToolFunction


def create_tools():
    """Create tools for the example."""
    return [
        Tool(
            type="function",
            function=ToolFunction(
                name="analyze_sentiment",
                description="Analyze the sentiment of a piece of text",
                parameters={
                    "type": "object",
                    "properties": {
                        "text": {
                            "type": "string",
                            "description": "The text to analyze",
                        },
                        "language": {
                            "type": "string",
                            "description": "The language of the text (optional)",
                        },
                    },
                    "required": ["text"],
                },
            ),
        ),
        Tool(
            type="function",
            function=ToolFunction(
                name="translate_text",
                description="Translate text from one language to another",
                parameters={
                    "type": "object",
                    "properties": {
                        "text": {
                            "type": "string",
                            "description": "The text to translate",
                        },
                        "source_language": {
                            "type": "string",
                            "description": "Source language code (e.g., 'en', 'es')",
                        },
                        "target_language": {
                            "type": "string",
                            "description": "Target language code",
                        },
                    },
                    "required": ["text", "target_language"],
                },
            ),
        ),
    ]


def execute_tool(name: str, args: Dict[str, Any]) -> Dict[str, Any]:
    """Simulate tool execution."""
    if name == "analyze_sentiment":
        text = args.get("text", "")
        # Simple mock sentiment analysis
        positive_words = ["good", "great", "happy", "love", "excellent"]
        negative_words = ["bad", "terrible", "sad", "hate", "awful"]

        text_lower = text.lower()
        pos_count = sum(1 for w in positive_words if w in text_lower)
        neg_count = sum(1 for w in negative_words if w in text_lower)

        if pos_count > neg_count:
            sentiment = "positive"
            score = 0.7
        elif neg_count > pos_count:
            sentiment = "negative"
            score = -0.7
        else:
            sentiment = "neutral"
            score = 0.0

        return {
            "text": text[:50] + "..." if len(text) > 50 else text,
            "sentiment": sentiment,
            "confidence": abs(score),
            "language": args.get("language", "en"),
        }
    elif name == "translate_text":
        # Mock translation
        return {
            "original": args.get("text"),
            "translated": f"[Translated to {args.get('target_language')}]: {args.get('text')}",
            "source_language": args.get("source_language", "auto-detected"),
            "target_language": args.get("target_language"),
        }
    return {"error": f"Unknown tool: {name}"}


def main():
    # Check for required environment
    project = os.getenv("GOOGLE_CLOUD_PROJECT")
    if not project:
        print("Please set GOOGLE_CLOUD_PROJECT environment variable")
        print("\nVertex AI requires:")
        print("  1. A Google Cloud project with Vertex AI API enabled")
        print("  2. Authentication via 'gcloud auth application-default login'")
        print("     or a service account key")
        print("  3. GOOGLE_CLOUD_PROJECT environment variable")
        return

    print("=== Vertex AI Tool Calling Example ===\n")
    print(f"Project: {project}")
    print(f"Location: {os.getenv('GOOGLE_CLOUD_LOCATION', 'us-central1')}\n")

    tools = create_tools()

    # Create model
    model = AIFactory.create_language(
        provider="vertex",
        model_name="gemini-2.0-flash",  # Vertex AI Gemini model
    )

    # Example 1: Basic tool call
    print("1. Basic Tool Call")
    print("-" * 40)
    messages = [
        {
            "role": "user",
            "content": "Analyze the sentiment of: 'I absolutely love this new product, it's excellent!'",
        }
    ]
    response = model.chat_complete(messages, tools=tools)

    if response.choices[0].message.tool_calls:
        tc = response.choices[0].message.tool_calls[0]
        print(f"Tool: {tc.function.name}")
        print(f"Args: {json.loads(tc.function.arguments)}")
    else:
        print(f"Response: {response.choices[0].message.content}")
    print()

    # Example 2: Multi-turn with tool results
    print("2. Multi-turn Conversation")
    print("-" * 40)

    messages = [
        {
            "role": "user",
            "content": "Translate 'Hello, how are you?' to Spanish",
        }
    ]

    response = model.chat_complete(messages, tools=tools)

    if response.choices[0].message.tool_calls:
        # Add assistant's tool call
        messages.append(
            {
                "role": "assistant",
                "tool_calls": [
                    {
                        "id": tc.id,
                        "type": tc.type,
                        "function": {
                            "name": tc.function.name,
                            "arguments": tc.function.arguments,
                        },
                    }
                    for tc in response.choices[0].message.tool_calls
                ],
            }
        )

        # Execute tools
        for tc in response.choices[0].message.tool_calls:
            args = json.loads(tc.function.arguments)
            result = execute_tool(tc.function.name, args)
            print(f"Executed: {tc.function.name}")
            print(f"Result: {result}")

            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tc.id,
                    "content": json.dumps(result),
                }
            )

        # Get final response
        final_response = model.chat_complete(messages, tools=tools)
        print(f"\nVertex AI response: {final_response.choices[0].message.content}")
    print()

    # Example 3: Chained tools
    print("3. Chained Tool Calls")
    print("-" * 40)
    messages = [
        {
            "role": "user",
            "content": "Translate 'The weather is terrible today' to French, then analyze its sentiment",
        }
    ]

    response = model.chat_complete(messages, tools=tools)

    iteration = 0
    while response.choices[0].message.tool_calls and iteration < 5:
        iteration += 1

        messages.append(
            {
                "role": "assistant",
                "tool_calls": [
                    {
                        "id": tc.id,
                        "type": tc.type,
                        "function": {
                            "name": tc.function.name,
                            "arguments": tc.function.arguments,
                        },
                    }
                    for tc in response.choices[0].message.tool_calls
                ],
            }
        )

        for tc in response.choices[0].message.tool_calls:
            args = json.loads(tc.function.arguments)
            result = execute_tool(tc.function.name, args)
            print(f"  Step {iteration}: {tc.function.name} -> {json.dumps(result)[:50]}...")

            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tc.id,
                    "content": json.dumps(result),
                }
            )

        response = model.chat_complete(messages, tools=tools)

    print(f"\nFinal response: {response.choices[0].message.content}")
    print()

    print("=== Example Complete ===")
    print("\nVertex AI notes:")
    print("  - Uses Google Cloud authentication (not API key)")
    print("  - Esperanto converts to Vertex AI SDK format internally")
    print("  - Tool call IDs are generated by Esperanto")


if __name__ == "__main__":
    main()



"""Azure OpenAI Text-to-Speech provider implementation."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import httpx

from .base import AudioResponse, Model, TextToSpeechModel, Voice


class AzureTextToSpeechModel(TextToSpeechModel):
    """Azure OpenAI Text-to-Speech implementation using direct HTTP.

    Supports Azure OpenAI TTS deployments with multiple voice options.
    Available voices: alloy, echo, fable, onyx, nova, shimmer
    """

    DEFAULT_MODEL = "tts-1"
    DEFAULT_VOICE = "alloy"
    AVAILABLE_VOICES = ["alloy", "echo", "fable", "onyx", "nova", "shimmer"]
    PROVIDER = "azure"

    def __init__(
        self,
        model_name: str = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs
    ):
        """Initialize Azure TTS provider.

        Args:
            model_name: Name of the Azure deployment to use
            api_key: Azure API key. If not provided, will try env vars
            base_url: Azure endpoint URL
            **kwargs: Additional configuration options
        """
        super().__init__(
            model_name=model_name,
            api_key=api_key,
            base_url=base_url,
            config=kwargs
        )

        # Resolve configuration with priority: config dict → modality env var → generic env var
        self.api_key = (
            self.api_key or
            self._config.get("api_key") or
            os.getenv("AZURE_OPENAI_API_KEY_TTS") or
            os.getenv("AZURE_OPENAI_API_KEY")
        )

        self.azure_endpoint = (
            self.base_url or
            self._config.get("azure_endpoint") or
            os.getenv("AZURE_OPENAI_ENDPOINT_TTS") or
            os.getenv("AZURE_OPENAI_ENDPOINT")
        )

        self.api_version = (
            self._config.get("api_version") or
            os.getenv("AZURE_OPENAI_API_VERSION_TTS") or
            os.getenv("AZURE_OPENAI_API_VERSION")
        )

        # deployment_name is model_name for Azure
        self.deployment_name = self.model_name or self._get_default_model()

        # Validate required parameters
        if not self.api_key:
            raise ValueError(
                "Azure OpenAI API key not found. Set AZURE_OPENAI_API_KEY_TTS "
                "or AZURE_OPENAI_API_KEY environment variable, or provide in config."
            )
        if not self.azure_endpoint:
            raise ValueError(
                "Azure OpenAI endpoint not found. Set AZURE_OPENAI_ENDPOINT_TTS "
                "or AZURE_OPENAI_ENDPOINT environment variable, or provide in config."
            )
        if not self.api_version:
            raise ValueError(
                "Azure OpenAI API version not found. Set AZURE_OPENAI_API_VERSION_TTS "
                "or AZURE_OPENAI_API_VERSION environment variable, or provide in config."
            )

        # Initialize HTTP clients with configurable timeout
        self._create_http_clients()

    def _get_headers(self) -> Dict[str, str]:
        """Get headers for Azure API requests."""
        return {
            "api-key": self.api_key,  # Azure uses api-key, not Bearer
            "Content-Type": "application/json",
        }

    def _build_url(self, path: str) -> str:
        """Build Azure OpenAI URL with deployment name."""
        # Remove trailing slash from endpoint
        endpoint = self.azure_endpoint.rstrip('/')
        # Azure URL pattern: {endpoint}/openai/deployments/{deployment}/{path}?api-version={version}
        return f"{endpoint}/openai/deployments/{self.deployment_name}/{path}?api-version={self.api_version}"

    def _handle_error(self, response: httpx.Response) -> None:
        """Handle HTTP error responses."""
        if response.status_code >= 400:
            try:
                error_data = response.json()
                error_message = error_data.get("error", {}).get("message", f"HTTP {response.status_code}")
            except Exception:
                error_message = f"HTTP {response.status_code}: {response.text}"
            raise RuntimeError(f"Azure OpenAI API error: {error_message}")

    @property
    def available_voices(self) -> Dict[str, Voice]:
        """Get available voices from Azure OpenAI TTS.

        Returns:
            Dict[str, Voice]: Dictionary of available voices with their information
        """
        voices = {
            "alloy": Voice(
                name="alloy",
                id="alloy",
                gender="NEUTRAL",
                language_code="en-US",
                description="Neutral and balanced voice"
            ),
            "echo": Voice(
                name="echo",
                id="echo",
                gender="MALE",
                language_code="en-US",
                description="Mature and deep voice"
            ),
            "fable": Voice(
                name="fable",
                id="fable",
                gender="FEMALE",
                language_code="en-US",
                description="Warm and expressive voice"
            ),
            "onyx": Voice(
                name="onyx",
                id="onyx",
                gender="MALE",
                language_code="en-US",
                description="Smooth and authoritative voice"
            ),
            "nova": Voice(
                name="nova",
                id="nova",
                gender="FEMALE",
                language_code="en-US",
                description="Energetic and bright voice"
            ),
            "shimmer": Voice(
                name="shimmer",
                id="shimmer",
                gender="FEMALE",
                language_code="en-US",
                description="Clear and professional voice"
            ),
        }
        return voices

    @property
    def provider(self) -> str:
        """Get the provider name."""
        return self.PROVIDER

    def _get_default_model(self) -> str:
        """Get the default model name."""
        return self.DEFAULT_MODEL

    def _get_models(self) -> List[Model]:
        """List all available models for this provider.

        Note: Azure doesn't have a models API endpoint - it uses deployments.
        Returns an empty list since model discovery isn't available.
        """
        return []

    def generate_speech(
        self,
        text: str,
        voice: str = "alloy",
        output_file: Optional[Union[str, Path]] = None,
        **kwargs
    ) -> AudioResponse:
        """Generate speech from text using Azure OpenAI TTS.

        Args:
            text: Text to convert to speech
            voice: Voice to use (default: "alloy")
            output_file: Optional path to save the audio file
            **kwargs: Additional parameters to pass to the Azure API

        Returns:
            AudioResponse containing the audio data and metadata

        Raises:
            RuntimeError: If speech generation fails
        """
        try:
            url = self._build_url("audio/speech")

            # Prepare request payload
            payload = {
                "model": self.deployment_name,
                "voice": voice,
                "input": text,
                **kwargs
            }

            # Generate speech
            response = self.client.post(
                url,
                headers=self._get_headers(),
                json=payload
            )
            self._handle_error(response)

            # Get audio data (binary content)
            audio_data = response.content

            # Save to file if specified
            if output_file:
                output_file = Path(output_file)
                output_file.parent.mkdir(parents=True, exist_ok=True)
                output_file.write_bytes(audio_data)

            return AudioResponse(
                audio_data=audio_data,
                content_type="audio/mp3",
                model=self.deployment_name,
                voice=voice,
                provider=self.PROVIDER,
                metadata={"text": text}
            )

        except Exception as e:
            raise RuntimeError(f"Failed to generate speech: {str(e)}") from e

    async def agenerate_speech(
        self,
        text: str,
        voice: str = "alloy",
        output_file: Optional[Union[str, Path]] = None,
        **kwargs
    ) -> AudioResponse:
        """Generate speech from text using Azure OpenAI TTS asynchronously.

        Args:
            text: Text to convert to speech
            voice: Voice to use (default: "alloy")
            output_file: Optional path to save the audio file
            **kwargs: Additional parameters to pass to the Azure API

        Returns:
            AudioResponse containing the audio data and metadata

        Raises:
            RuntimeError: If speech generation fails
        """
        try:
            url = self._build_url("audio/speech")

            # Prepare request payload
            payload = {
                "model": self.deployment_name,
                "voice": voice,
                "input": text,
                **kwargs
            }

            # Generate speech
            response = await self.async_client.post(
                url,
                headers=self._get_headers(),
                json=payload
            )
            self._handle_error(response)

            # Get audio data (binary content)
            audio_data = response.content

            # Save to file if specified
            if output_file:
                output_file = Path(output_file)
                output_file.parent.mkdir(parents=True, exist_ok=True)
                output_file.write_bytes(audio_data)

            return AudioResponse(
                audio_data=audio_data,
                content_type="audio/mp3",
                model=self.deployment_name,
                voice=voice,
                provider=self.PROVIDER,
                metadata={"text": text}
            )

        except Exception as e:
            raise RuntimeError(f"Failed to generate speech: {str(e)}") from e



"""Reranker types for Esperanto."""

from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field


class RerankResult(BaseModel):
    """Individual reranking result for a document."""

    index: int = Field(description="Original document index", ge=0)
    document: str = Field(description="Original document text")
    relevance_score: float = Field(
        description="Normalized 0-1 relevance score", ge=0.0, le=1.0
    )

    model_config = ConfigDict(frozen=True)


class RerankResponse(BaseModel):
    """Standardized reranking response across all providers."""

    results: List[RerankResult] = Field(
        description="Ranked results (highest score first)"
    )
    model: str = Field(description="Model used for reranking")
    usage: Optional["Usage"] = Field(
        default=None, description="Token/request usage if available"
    )

    @property
    def top_result(self) -> Optional[RerankResult]:
        """Get the highest ranked result."""
        return self.results[0] if self.results else None

    def get_top_k(self, k: int) -> List[RerankResult]:
        """Get top k results."""
        return self.results[:k]

    model_config = ConfigDict(frozen=True)


# Import Usage after defining our classes to avoid circular imports
from .response import Usage

# Update forward reference
RerankResponse.model_rebuild()


"""Jina AI embedding model implementation."""

import os
from typing import Any, Dict, List

import httpx

from esperanto.common_types import Model
from esperanto.common_types.task_type import EmbeddingTaskType

from .base import EmbeddingModel


class JinaEmbeddingModel(EmbeddingModel):
    """Jina AI embeddings with native support for all advanced features."""

    # Jina supports all advanced features natively
    SUPPORTED_FEATURES = ["task_type", "late_chunking", "output_dimensions", "truncate_at_max_length"]

    # Task type mapping from universal enum to Jina API values
    TASK_MAPPING = {
        EmbeddingTaskType.RETRIEVAL_QUERY: "retrieval.query",
        EmbeddingTaskType.RETRIEVAL_DOCUMENT: "retrieval.passage",
        EmbeddingTaskType.SIMILARITY: "text-matching",
        EmbeddingTaskType.CLASSIFICATION: "classification",
        EmbeddingTaskType.CLUSTERING: "separation",
        EmbeddingTaskType.CODE_RETRIEVAL: "code.query",
        EmbeddingTaskType.DEFAULT: None
    }

    def __init__(self, **kwargs):
        """Initialize Jina embedding model.

        Args:
            **kwargs: Configuration parameters including:
                - api_key: Jina API key (or set JINA_API_KEY env var)
                - model_name: Model to use (default: jina-embeddings-v3)
                - base_url: API base URL (default: https://api.jina.ai/v1/embeddings)
                - config: Dict with task_type, late_chunking, output_dimensions, etc.
        """
        super().__init__(**kwargs)
        self.api_key = kwargs.get("api_key") or os.getenv("JINA_API_KEY")
        if not self.api_key:
            raise ValueError(
                "Jina API key not found. Please set the JINA_API_KEY environment "
                "variable or pass it as 'api_key' parameter."
            )
        self.base_url = kwargs.get("base_url", "https://api.jina.ai/v1/embeddings")

        # Initialize HTTP clients with configurable timeout
        self._create_http_clients()

    def _apply_task_optimization(self, texts: List[str]) -> List[str]:
        """Jina handles task optimization natively via API."""
        # Don't apply prefix - Jina API handles this natively
        return texts

    def _apply_late_chunking(self, texts: List[str]) -> List[str]:
        """Jina handles late chunking natively via API."""
        # Don't chunk here - Jina API handles this natively
        return texts

    def _get_headers(self) -> Dict[str, str]:
        """Get request headers."""
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json"
        }

    def _map_task_type(self) -> str:
        """Map universal task type to Jina-specific value."""
        if not self.task_type:
            return None
        return self.TASK_MAPPING.get(self.task_type)

    def _build_request_payload(self, texts: List[str]) -> Dict[str, Any]:
        """Build request payload for Jina API.

        Args:
            texts: List of texts to embed.

        Returns:
            Request payload dict.
        """
        # Clean texts using base class method
        cleaned_texts = [self._clean_text(text) for text in texts]

        # Build Jina-specific payload
        input_data = [{"text": text} for text in cleaned_texts]

        payload = {
            "model": self.get_model_name(),
            "input": input_data
        }

        # Map and add task type if specified
        if self.task_type:
            jina_task = self._map_task_type()
            if jina_task:
                payload["task"] = jina_task

        # Add other advanced features - all natively supported
        if self.late_chunking:
            payload["late_chunking"] = True

        if self.truncate_at_max_length:
            payload["truncate"] = True

        if self.output_dimensions:
            payload["dimensions"] = self.output_dimensions

        return payload

    def _handle_error(self, response: httpx.Response) -> None:
        """Handle error responses from Jina API.

        Args:
            response: HTTP response object.

        Raises:
            RuntimeError: With details from the error response.
        """
        try:
            error_data = response.json()
            error_message = error_data.get("error", {}).get("message", "Unknown error")
            error_type = error_data.get("error", {}).get("type", "Unknown")
            raise RuntimeError(f"Jina API error ({error_type}): {error_message}")
        except (KeyError, ValueError):
            raise RuntimeError(f"Jina API error: {response.status_code} - {response.text}")

    def embed(self, texts: List[str], **kwargs) -> List[List[float]]:
        """Create embeddings for the given texts.

        Args:
            texts: List of texts to create embeddings for.
            **kwargs: Additional arguments (not used for Jina).

        Returns:
            List of embeddings, one for each input text.
        """
        if not texts:
            return []

        payload = self._build_request_payload(texts)

        try:
            response = self.client.post(
                self.base_url,
                json=payload,
                headers=self._get_headers()
            )

            if response.status_code != 200:
                self._handle_error(response)

            response_data = response.json()

            # Extract embeddings from response
            embeddings = []
            for item in response_data.get("data", []):
                embedding = item.get("embedding")
                if embedding:
                    # Ensure all values are floats for consistency
                    embeddings.append([float(value) for value in embedding])

            return embeddings

        except httpx.TimeoutException:
            raise RuntimeError("Request to Jina API timed out")
        except httpx.RequestError as e:
            raise RuntimeError(f"Network error calling Jina API: {str(e)}")
        finally:
            pass  # Client is reused, don't close

    async def aembed(self, texts: List[str], **kwargs) -> List[List[float]]:
        """Create embeddings for the given texts asynchronously.

        Args:
            texts: List of texts to create embeddings for.
            **kwargs: Additional arguments (not used for Jina).

        Returns:
            List of embeddings, one for each input text.
        """
        if not texts:
            return []

        payload = self._build_request_payload(texts)

        try:
            response = await self.async_client.post(
                self.base_url,
                json=payload,
                headers=self._get_headers()
            )

            if response.status_code != 200:
                self._handle_error(response)

            response_data = response.json()

            # Extract embeddings from response
            embeddings = []
            for item in response_data.get("data", []):
                embedding = item.get("embedding")
                if embedding:
                    # Ensure all values are floats for consistency
                    embeddings.append([float(value) for value in embedding])

            return embeddings

        except httpx.TimeoutException:
            raise RuntimeError("Request to Jina API timed out")
        except httpx.RequestError as e:
            raise RuntimeError(f"Network error calling Jina API: {str(e)}")

    @property
    def provider(self) -> str:
        """Get the provider name."""
        return "jina"

    def _get_models(self) -> List[Model]:
        """List available Jina embedding models."""
        return [
            Model(
                id="jina-embeddings-v4",
                owned_by="jina",
                context_window=8192),
            Model(
                id="jina-embeddings-v3",
                owned_by="jina",
                context_window=8192),
            Model(
                id="jina-embeddings-v2-base-en",
                owned_by="jina",
                context_window=8192),
            Model(
                id="jina-embeddings-v2-small-en",
                owned_by="jina",
                context_window=8192),
            Model(
                id="jina-embeddings-v2-base-multilingual",
                owned_by="jina",
                context_window=8192),
            Model(
                id="jina-clip-v1",
                owned_by="jina",
                context_window=None,  # Multimodal model
            ),
            Model(
                id="jina-clip-v2",
                owned_by="jina",
                context_window=None,  # Multimodal model
            ),
        ]

    def _get_default_model(self) -> str:
        """Get the default model name."""
        return "jina-embeddings-v3"


"""Tests for the Perplexity AI language model provider."""

import os
from unittest.mock import AsyncMock, Mock, patch

import pytest
from langchain_openai import ChatOpenAI

from esperanto.providers.llm.perplexity import PerplexityLanguageModel
from esperanto.common_types import (
    ChatCompletion, Choice, Message, Usage,
    Tool, ToolFunction, ToolCall, ToolCallValidationError
)


@pytest.fixture
def mock_httpx_response():
    """Mock httpx response for Perplexity API."""
    def create_response():
        return {
            "id": "cmpl-123",
            "object": "chat.completion",
            "created": 123,
            "model": "llama-3-sonar-large-32k-online",
            "choices": [
                {
                    "index": 0,
                    "message": {
                        "role": "assistant",
                        "content": "Hello!"
                    },
                    "finish_reason": "stop"
                }
            ],
            "usage": {
                "prompt_tokens": 2,
                "completion_tokens": 3,
                "total_tokens": 5
            }
        }
    return create_response

@pytest.fixture
def perplexity_provider(mock_httpx_response):
    """Fixture for PerplexityLanguageModel."""
    # Set dummy API key for testing
    os.environ["PERPLEXITY_API_KEY"] = "test_api_key"
    provider = PerplexityLanguageModel(model_name="llama-3-sonar-large-32k-online")
    
    # Mock the HTTP clients
    mock_client = Mock()
    mock_async_client = AsyncMock()
    
    def mock_post(url, **kwargs):
        response_data = mock_httpx_response()
        
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = response_data
        return mock_response
    
    async def mock_async_post(url, **kwargs):
        response_data = mock_httpx_response()
        
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = response_data
        return mock_response
    
    mock_client.post = mock_post
    mock_async_client.post = mock_async_post
    
    provider.client = mock_client
    provider.async_client = mock_async_client
    
    # Clean up env var after test
    yield provider
    del os.environ["PERPLEXITY_API_KEY"]




def test_perplexity_provider_initialization(perplexity_provider):
    """Test initialization of PerplexityLanguageModel."""
    assert perplexity_provider.provider == "perplexity"
    assert (
        perplexity_provider.get_model_name() == "llama-3-sonar-large-32k-online"
    )  # Default model
    assert perplexity_provider.api_key == "test_api_key"
    assert perplexity_provider.base_url == "https://api.perplexity.ai"


def test_perplexity_provider_initialization_no_api_key():
    """Test initialization raises error if API key is missing."""
    if "PERPLEXITY_API_KEY" in os.environ:
        del os.environ["PERPLEXITY_API_KEY"]  # Ensure key is not set
    with pytest.raises(ValueError, match="Perplexity API key not found"):
        PerplexityLanguageModel(model_name="test-model")


def test_perplexity_get_api_kwargs(perplexity_provider):
    """Test _get_api_kwargs includes standard and perplexity-specific args."""
    perplexity_provider.temperature = 0.8
    perplexity_provider.max_tokens = 500
    perplexity_provider.search_domain_filter = ["example.com"]
    perplexity_provider.return_images = True
    perplexity_provider.web_search_options = {"search_context_size": "medium"}

    kwargs = perplexity_provider._get_api_kwargs()
    perplexity_params = perplexity_provider._get_perplexity_params()

    # Test standard kwargs
    assert kwargs["temperature"] == 0.8
    assert kwargs["max_tokens"] == 500
    # Ensure Perplexity params are NOT in standard kwargs
    assert "search_domain_filter" not in kwargs
    assert "return_images" not in kwargs
    assert "web_search_options" not in kwargs

    # Test perplexity params
    assert perplexity_params["search_domain_filter"] == ["example.com"]
    assert perplexity_params["return_images"] is True
    assert "return_related_questions" not in perplexity_params  # Not set
    assert "search_recency_filter" not in perplexity_params  # Not set
    assert perplexity_params["web_search_options"] == {"search_context_size": "medium"}


def test_perplexity_get_api_kwargs_exclude_stream(perplexity_provider):
    """Test _get_api_kwargs excludes stream when requested."""
    perplexity_provider.streaming = True
    kwargs = perplexity_provider._get_api_kwargs(exclude_stream=True)
    assert "stream" not in kwargs


@pytest.mark.asyncio
async def test_perplexity_async_call(perplexity_provider):
    """Test the asynchronous call method."""
    # Pass messages as dicts, not LangChain objects
    messages = [{"role": "user", "content": "Hello"}]
    expected_response_text = "Hello!"

    response = await perplexity_provider.achat_complete(messages)

    assert response.choices[0].message.content == expected_response_text
    assert response.model == perplexity_provider.get_model_name()


def test_perplexity_call(perplexity_provider):
    """Test the synchronous call method."""
    # Pass messages as dicts, not LangChain objects
    messages = [
        {"role": "system", "content": "Be brief."},
        {"role": "user", "content": "Hi"},
    ]
    expected_response_text = "Hello!"

    response = perplexity_provider.chat_complete(messages)

    assert response.choices[0].message.content == expected_response_text
    assert response.model == perplexity_provider.get_model_name()


def test_perplexity_call_with_extra_params(perplexity_provider):
    """Test synchronous call with extra Perplexity parameters."""
    perplexity_provider.search_domain_filter = ["test.com"]
    perplexity_provider.return_images = True
    messages = [{"role": "user", "content": "Hi"}]
    expected_response_text = "Hello!"

    response = perplexity_provider.chat_complete(messages)

    assert response.choices[0].message.content == expected_response_text
    assert response.model == perplexity_provider.get_model_name()
    
    # Test that perplexity params are available
    params = perplexity_provider._get_perplexity_params()
    assert params["search_domain_filter"] == ["test.com"]
    assert params["return_images"] is True


def test_perplexity_to_langchain(perplexity_provider):
    """Test conversion to LangChain model."""
    perplexity_provider.temperature = 0.7
    perplexity_provider.max_tokens = 100
    perplexity_provider.search_domain_filter = ["test.dev"]
    perplexity_provider.return_related_questions = True

    langchain_model = perplexity_provider.to_langchain()

    assert isinstance(langchain_model, ChatOpenAI)
    assert langchain_model.model_name == perplexity_provider.get_model_name()
    # Skip API key and base_url checks since they may be private attributes in LangChain
    assert langchain_model.temperature == 0.7
    assert langchain_model.max_tokens == 100
    assert langchain_model.model_kwargs["search_domain_filter"] == ["test.dev"]
    assert langchain_model.model_kwargs["return_related_questions"] is True
    assert "return_images" not in langchain_model.model_kwargs  # Not set


def test_perplexity_to_langchain_structured(perplexity_provider):
    """Test conversion to LangChain model with structured output."""
    perplexity_provider.structured = {"type": "json_object"}
    langchain_model = perplexity_provider.to_langchain()

    assert langchain_model.model_kwargs["response_format"] == {"type": "text"}


def test_perplexity_models_property(perplexity_provider):
    """Test the models property (currently hardcoded)."""
    models = perplexity_provider.models
    assert isinstance(models, list)
    assert len(models) > 5  # Check if it returns a reasonable number of models
    assert all(model.owned_by == "Perplexity" for model in models)
    # Check for some known models
    model_ids = [m.id for m in models]
    assert "sonar-pro" in model_ids
    assert "sonar" in model_ids


# =============================================================================
# Tool Calling Tests
# =============================================================================


@pytest.fixture
def sample_tools():
    """Sample tools for testing."""
    return [
        Tool(
            function=ToolFunction(
                name="get_weather",
                description="Get the current weather for a location",
                parameters={
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city name"},
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]}
                    },
                    "required": ["location"]
                }
            )
        ),
        Tool(
            function=ToolFunction(
                name="get_time",
                description="Get the current time for a timezone",
                parameters={
                    "type": "object",
                    "properties": {
                        "timezone": {"type": "string", "description": "The timezone"}
                    }
                }
            )
        )
    ]


@pytest.fixture
def mock_perplexity_tool_call_response():
    """Mock HTTP response for Perplexity chat completions with tool calls."""
    return {
        "id": "chatcmpl-tool-123",
        "object": "chat.completion",
        "created": 1677652288,
        "model": "llama-3-sonar-large-32k-online",
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": None,
                    "tool_calls": [
                        {
                            "id": "call_abc123",
                            "type": "function",
                            "function": {
                                "name": "get_weather",
                                "arguments": '{"location": "San Francisco", "unit": "celsius"}'
                            }
                        }
                    ]
                },
                "finish_reason": "tool_calls"
            }
        ],
        "usage": {
            "prompt_tokens": 50,
            "completion_tokens": 20,
            "total_tokens": 70
        }
    }


@pytest.fixture
def perplexity_model_with_tool_response(mock_perplexity_tool_call_response):
    """Create a Perplexity model with tool call response mocked."""
    os.environ["PERPLEXITY_API_KEY"] = "test_api_key"
    model = PerplexityLanguageModel(model_name="llama-3-sonar-large-32k-online")

    mock_client = Mock()
    mock_async_client = AsyncMock()

    def make_response(status_code, json_data):
        response = Mock()
        response.status_code = status_code
        response.json.return_value = json_data
        return response

    mock_client.post = Mock(return_value=make_response(200, mock_perplexity_tool_call_response))
    mock_async_client.post = AsyncMock(return_value=make_response(200, mock_perplexity_tool_call_response))

    model.client = mock_client
    model.async_client = mock_async_client
    yield model
    del os.environ["PERPLEXITY_API_KEY"]


class TestToolConversion:
    """Tests for tool conversion to OpenAI format (Perplexity uses OpenAI-compatible format)."""

    def test_convert_single_tool(self, perplexity_provider, sample_tools):
        """Test converting a single tool to OpenAI format."""
        result = perplexity_provider._convert_tools_to_openai([sample_tools[0]])

        assert len(result) == 1
        assert result[0]["type"] == "function"
        assert result[0]["function"]["name"] == "get_weather"
        assert result[0]["function"]["description"] == "Get the current weather for a location"
        assert result[0]["function"]["parameters"]["type"] == "object"
        assert "location" in result[0]["function"]["parameters"]["properties"]

    def test_convert_multiple_tools(self, perplexity_provider, sample_tools):
        """Test converting multiple tools to OpenAI format."""
        result = perplexity_provider._convert_tools_to_openai(sample_tools)

        assert len(result) == 2
        assert result[0]["function"]["name"] == "get_weather"
        assert result[1]["function"]["name"] == "get_time"

    def test_convert_none_tools(self, perplexity_provider):
        """Test converting None returns None."""
        result = perplexity_provider._convert_tools_to_openai(None)
        assert result is None

    def test_convert_empty_tools(self, perplexity_provider):
        """Test converting empty list returns None."""
        result = perplexity_provider._convert_tools_to_openai([])
        assert result is None


class TestToolCallResponse:
    """Tests for handling tool call responses."""

    def test_chat_complete_with_tools(self, perplexity_model_with_tool_response, sample_tools):
        """Test chat_complete with tools returns tool calls."""
        messages = [{"role": "user", "content": "What's the weather in SF?"}]

        response = perplexity_model_with_tool_response.chat_complete(
            messages, tools=sample_tools
        )

        # Check payload included tools
        call_args = perplexity_model_with_tool_response.client.post.call_args
        json_payload = call_args[1]["json"]
        assert "tools" in json_payload
        assert len(json_payload["tools"]) == 2

        # Check response has tool calls
        assert len(response.choices) == 1
        assert response.choices[0].message.tool_calls is not None
        assert len(response.choices[0].message.tool_calls) == 1

        tool_call = response.choices[0].message.tool_calls[0]
        assert isinstance(tool_call, ToolCall)
        assert tool_call.id == "call_abc123"
        assert tool_call.function.name == "get_weather"
        assert '"location": "San Francisco"' in tool_call.function.arguments

    def test_chat_complete_with_tool_choice(self, perplexity_model_with_tool_response, sample_tools):
        """Test chat_complete with tool_choice parameter."""
        messages = [{"role": "user", "content": "What's the weather?"}]

        perplexity_model_with_tool_response.chat_complete(
            messages, tools=sample_tools, tool_choice="required"
        )

        call_args = perplexity_model_with_tool_response.client.post.call_args
        json_payload = call_args[1]["json"]
        assert json_payload["tool_choice"] == "required"

    @pytest.mark.asyncio
    async def test_achat_complete_with_tools(self, perplexity_model_with_tool_response, sample_tools):
        """Test async chat_complete with tools returns tool calls."""
        messages = [{"role": "user", "content": "What's the weather in SF?"}]

        response = await perplexity_model_with_tool_response.achat_complete(
            messages, tools=sample_tools
        )

        # Check payload included tools
        call_args = perplexity_model_with_tool_response.async_client.post.call_args
        json_payload = call_args[1]["json"]
        assert "tools" in json_payload

        # Check response has tool calls
        assert response.choices[0].message.tool_calls is not None
        tool_call = response.choices[0].message.tool_calls[0]
        assert tool_call.function.name == "get_weather"


class TestToolCallValidation:
    """Tests for tool call validation."""

    def test_validation_passes_for_valid_tool_call(
        self, perplexity_model_with_tool_response, sample_tools
    ):
        """Test that validation passes for valid tool calls."""
        pytest.importorskip("jsonschema")

        messages = [{"role": "user", "content": "What's the weather in SF?"}]

        # Should not raise
        response = perplexity_model_with_tool_response.chat_complete(
            messages, tools=sample_tools, validate_tool_calls=True
        )

        assert response.choices[0].message.tool_calls is not None



"""SSL verification configuration utilities for Esperanto providers."""

import os
import warnings
from typing import Any, Union

# Environment variable names for SSL configuration
SSL_VERIFY_ENV_VAR = "ESPERANTO_SSL_VERIFY"
SSL_CA_BUNDLE_ENV_VAR = "ESPERANTO_SSL_CA_BUNDLE"


class SSLMixin:
    """Mixin providing SSL verification configuration functionality.

    This mixin provides a standardized way to configure SSL verification across all
    Esperanto providers. It implements a priority system:

    1. Config dict ssl_ca_bundle (highest priority)
    2. Config dict verify_ssl
    3. Environment variable ESPERANTO_SSL_CA_BUNDLE
    4. Environment variable ESPERANTO_SSL_VERIFY
    5. Default: True (SSL verification enabled)

    The mixin must be used with classes that have:
    - _config: Dict[str, Any] attribute

    Example:
        # Disable SSL verification (development only)
        model = AIFactory.create_language(
            "ollama",
            "llama3",
            config={"verify_ssl": False}
        )

        # Use custom CA bundle (recommended for self-signed certs)
        model = AIFactory.create_language(
            "ollama",
            "llama3",
            config={"ssl_ca_bundle": "/path/to/ca-bundle.pem"}
        )
    """

    def _get_ssl_verify(self) -> Union[bool, str]:
        """Get SSL verification setting using priority hierarchy.

        Priority order (highest to lowest):
        1. Config dict ssl_ca_bundle: config={"ssl_ca_bundle": "/path/to/ca.pem"}
        2. Config dict verify_ssl: config={"verify_ssl": False}
        3. Environment variable: ESPERANTO_SSL_CA_BUNDLE=/path/to/ca.pem
        4. Environment variable: ESPERANTO_SSL_VERIFY=false
        5. Default: True (SSL verification enabled)

        Returns:
            Union[bool, str]:
                - True: SSL verification enabled using default CA bundle
                - False: SSL verification disabled (insecure)
                - str: Path to custom CA bundle file

        Raises:
            ValueError: If CA bundle path does not exist or verify_ssl has invalid type
        """
        # 1. Config dict ssl_ca_bundle (highest priority)
        if hasattr(self, "_config"):
            ca_bundle = self._config.get("ssl_ca_bundle")
            if ca_bundle:
                self._validate_ca_bundle(ca_bundle)
                return ca_bundle

            # 2. Config dict verify_ssl
            verify_ssl = self._config.get("verify_ssl")
            if verify_ssl is not None:
                validated = self._validate_verify_ssl(verify_ssl)
                if not validated:
                    self._emit_ssl_warning()
                return validated

        # 3. Environment variable for CA bundle
        ca_bundle_env = os.getenv(SSL_CA_BUNDLE_ENV_VAR)
        if ca_bundle_env:
            self._validate_ca_bundle(ca_bundle_env)
            return ca_bundle_env

        # 4. Environment variable for verify_ssl
        verify_env = os.getenv(SSL_VERIFY_ENV_VAR, "").lower()
        if verify_env in ("false", "0", "no"):
            self._emit_ssl_warning()
            return False

        # 5. Default: SSL verification enabled
        return True

    def _validate_verify_ssl(self, value: Any) -> bool:
        """Validate and normalize verify_ssl configuration value.

        Accepts boolean values directly, and converts common string representations
        of boolean values. Raises ValueError for invalid types.

        Args:
            value: The verify_ssl configuration value to validate

        Returns:
            bool: Normalized boolean value

        Raises:
            ValueError: If value cannot be converted to boolean
        """
        if isinstance(value, bool):
            return value
        if isinstance(value, (int, float)):
            return bool(value)
        if isinstance(value, str):
            if value.lower() in ("false", "0", "no"):
                return False
            if value.lower() in ("true", "1", "yes"):
                return True
        raise ValueError(
            f"verify_ssl must be a boolean, got {type(value).__name__}: {value!r}"
        )

    def _validate_ca_bundle(self, path: str) -> None:
        """Validate that CA bundle file exists.

        Args:
            path: Path to CA bundle file

        Raises:
            ValueError: If file does not exist
        """
        if not os.path.isfile(path):
            raise ValueError(f"CA bundle file not found: {path}")

    def _emit_ssl_warning(self) -> None:
        """Emit a security warning when SSL verification is disabled.

        This warning helps users identify when they're running in an insecure
        mode, which is particularly important if disabled configuration
        accidentally reaches production.
        """
        warnings.warn(
            "SSL verification is disabled. This is insecure and should only be used "
            "in development/testing environments. For production, use ssl_ca_bundle "
            "to specify a custom CA certificate instead.",
            UserWarning,
            stacklevel=4,
        )



"""Google GenAI speech-to-text provider implementation."""

import base64
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, Union

import httpx

from esperanto.common_types import TranscriptionResponse
from esperanto.providers.stt.base import Model, SpeechToTextModel


@dataclass
class GoogleSpeechToTextModel(SpeechToTextModel):
    """Google GenAI speech-to-text model implementation.

    Uses the Gemini API for audio transcription via the generateContent endpoint.
    This is NOT Cloud Speech-to-Text API v2 Chirp 3, but Gemini's audio understanding.
    """

    def __post_init__(self):
        """Initialize HTTP clients."""
        # Call parent's post_init to handle config initialization
        super().__post_init__()

        # Get API key - check both GOOGLE_API_KEY and GEMINI_API_KEY
        self._api_key = (
            self.api_key or os.getenv("GOOGLE_API_KEY") or os.getenv("GEMINI_API_KEY")
        )
        if not self._api_key:
            raise ValueError(
                "Google API key not found. Please set GOOGLE_API_KEY or GEMINI_API_KEY environment variable."
            )

        # Set base URL - consistent with other Google providers
        base_host = os.getenv("GEMINI_API_BASE_URL") or "https://generativelanguage.googleapis.com"
        self.base_url = f"{base_host}/v1beta"

        # Initialize HTTP clients with configurable timeout
        self._create_http_clients()

    def _get_default_model(self) -> str:
        """Get the default model name."""
        return "gemini-2.5-flash"

    @property
    def provider(self) -> str:
        """Get the provider name."""
        return "google"

    def _get_models(self) -> List[Model]:
        """List all available models for this provider.

        Returns hardcoded list of Gemini models that support audio transcription.
        """
        return [
            Model(
                id="gemini-2.5-flash",
                owned_by="Google",
                context_window=1000000,
            ),
            Model(
                id="gemini-2.0-flash",
                owned_by="Google",
                context_window=1000000,
            ),
        ]

    def _get_headers(self) -> Dict[str, str]:
        """Get headers for Google API requests."""
        return {
            "Content-Type": "application/json",
        }

    def _handle_error(self, response: httpx.Response) -> None:
        """Handle HTTP error responses."""
        if response.status_code >= 400:
            try:
                error_data = response.json()
                error_message = error_data.get("error", {}).get("message", f"HTTP {response.status_code}")
            except Exception:
                error_message = f"HTTP {response.status_code}: {response.text}"
            raise RuntimeError(f"Google API error: {error_message}")

    def _get_mime_type(self, file_path: str) -> str:
        """Get MIME type from file extension.

        Args:
            file_path: Path to audio file

        Returns:
            MIME type string

        Raises:
            ValueError: If file extension is not supported
        """
        extension = Path(file_path).suffix.lower()
        mime_types = {
            ".mp3": "audio/mp3",
            ".wav": "audio/wav",
            ".aiff": "audio/aiff",
            ".aac": "audio/aac",
            ".ogg": "audio/ogg",
            ".flac": "audio/flac",
        }

        if extension not in mime_types:
            raise ValueError(
                f"Unsupported audio format: {extension}. "
                f"Supported formats: {', '.join(mime_types.keys())}"
            )

        return mime_types[extension]

    def _encode_audio(self, audio_file: Union[str, BinaryIO]) -> Tuple[bytes, str]:
        """Read and base64 encode audio file.

        Args:
            audio_file: Path to audio file or file-like object

        Returns:
            Tuple of (audio_bytes, file_path_or_name)
        """
        if isinstance(audio_file, str):
            # File path
            with open(audio_file, "rb") as f:
                audio_bytes = f.read()
            return audio_bytes, audio_file
        else:
            # BinaryIO
            audio_bytes = audio_file.read()
            filename = getattr(audio_file, 'name', 'audio_input.mp3')
            return audio_bytes, filename

    def _build_prompt(
        self,
        language: Optional[str] = None,
        prompt: Optional[str] = None
    ) -> str:
        """Build transcription prompt for Gemini API.

        Args:
            language: Optional language code to hint the transcription
            prompt: Optional user-provided prompt to guide transcription

        Returns:
            Complete prompt string
        """
        base_prompt = "Generate a transcript of the speech in this audio file."

        parts = [base_prompt]

        if language:
            parts.append(f"The audio is in {language} language.")

        if prompt:
            parts.append(prompt)

        return " ".join(parts)

    def _parse_response(self, response_data: Dict[str, Any]) -> str:
        """Parse transcription text from Gemini API response.

        Args:
            response_data: JSON response from Gemini API

        Returns:
            Transcribed text

        Raises:
            RuntimeError: If response structure is unexpected
        """
        try:
            text = response_data["candidates"][0]["content"]["parts"][0]["text"]
            return text
        except (KeyError, IndexError) as e:
            raise RuntimeError(
                f"Failed to parse transcription from response: {e}. "
                f"Response structure: {list(response_data.keys())}"
            )

    def _prepare_request_payload(
        self,
        audio_file: Union[str, BinaryIO],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
    ) -> Tuple[Dict[str, Any], str]:
        """Prepare request payload for transcription.

        Args:
            audio_file: Path to audio file or file-like object
            language: Optional language code
            prompt: Optional text to guide transcription

        Returns:
            Tuple of (payload dict, model_name)
        """
        # Encode audio
        audio_bytes, file_identifier = self._encode_audio(audio_file)
        audio_base64 = base64.b64encode(audio_bytes).decode("utf-8")

        # Get MIME type
        mime_type = self._get_mime_type(file_identifier)

        # Build prompt
        text_prompt = self._build_prompt(language, prompt)

        # Construct request payload
        model_name = self.get_model_name()
        payload = {
            "contents": [{
                "parts": [
                    {"text": text_prompt},
                    {
                        "inline_data": {
                            "mime_type": mime_type,
                            "data": audio_base64
                        }
                    }
                ]
            }]
        }

        return payload, model_name

    def transcribe(
        self,
        audio_file: Union[str, BinaryIO],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
    ) -> TranscriptionResponse:
        """Transcribe audio to text using Google Gemini API.

        Args:
            audio_file: Path to audio file or file-like object
            language: Optional language code (e.g., 'en', 'es', 'pt')
            prompt: Optional text to guide the transcription

        Returns:
            TranscriptionResponse containing the transcribed text and metadata
        """
        # Prepare request
        payload, model_name = self._prepare_request_payload(audio_file, language, prompt)

        # Make HTTP request
        response = self.client.post(
            f"{self.base_url}/models/{model_name}:generateContent?key={self._api_key}",
            headers=self._get_headers(),
            json=payload
        )

        # Handle errors
        self._handle_error(response)

        # Parse response
        response_data = response.json()
        text = self._parse_response(response_data)

        return TranscriptionResponse(
            text=text,
            language=language,  # Pass through from request, not detected
            model=model_name,
        )

    async def atranscribe(
        self,
        audio_file: Union[str, BinaryIO],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
    ) -> TranscriptionResponse:
        """Async transcribe audio to text using Google Gemini API.

        Args:
            audio_file: Path to audio file or file-like object
            language: Optional language code (e.g., 'en', 'es', 'pt')
            prompt: Optional text to guide the transcription

        Returns:
            TranscriptionResponse containing the transcribed text and metadata
        """
        # Prepare request
        payload, model_name = self._prepare_request_payload(audio_file, language, prompt)

        # Make async HTTP request
        response = await self.async_client.post(
            f"{self.base_url}/models/{model_name}:generateContent?key={self._api_key}",
            headers=self._get_headers(),
            json=payload
        )

        # Handle errors
        self._handle_error(response)

        # Parse response
        response_data = response.json()
        text = self._parse_response(response_data)

        return TranscriptionResponse(
            text=text,
            language=language,  # Pass through from request, not detected
            model=model_name,
        )



"""OpenAI speech-to-text provider."""

import os
from dataclasses import dataclass
from typing import Any, BinaryIO, Dict, List, Optional, Union

import httpx

from esperanto.common_types import TranscriptionResponse
from esperanto.providers.stt.base import Model, SpeechToTextModel


@dataclass
class OpenAISpeechToTextModel(SpeechToTextModel):
    """OpenAI speech-to-text model implementation."""

    def __post_init__(self):
        """Initialize HTTP clients."""
        # Call parent's post_init to handle config initialization
        super().__post_init__()

        # Get API key
        self.api_key = self.api_key or os.getenv("OPENAI_API_KEY")
        if not self.api_key:
            raise ValueError("OpenAI API key not found")

        # Set base URL
        self.base_url = self.base_url or "https://api.openai.com/v1"

        # Initialize HTTP clients with configurable timeout
        self._create_http_clients()

    def _get_headers(self) -> Dict[str, str]:
        """Get headers for OpenAI API requests."""
        headers = {
            "Authorization": f"Bearer {self.api_key}",
        }
        # Organization is optional for STT models
        if hasattr(self, 'organization') and self.organization:
            headers["OpenAI-Organization"] = self.organization
        return headers

    def _handle_error(self, response: httpx.Response) -> None:
        """Handle HTTP error responses."""
        if response.status_code >= 400:
            try:
                error_data = response.json()
                error_message = error_data.get("error", {}).get("message", f"HTTP {response.status_code}")
            except Exception:
                error_message = f"HTTP {response.status_code}: {response.text}"
            raise RuntimeError(f"OpenAI API error: {error_message}")

    def _get_default_model(self) -> str:
        """Get the default model name."""
        return "whisper-1"

    @property
    def provider(self) -> str:
        """Get the provider name."""
        return "openai"

    def _get_models(self) -> List[Model]:
        """List all available models for this provider."""
        try:
            response = self.client.get(
                f"{self.base_url}/models",
                headers=self._get_headers()
            )
            self._handle_error(response)
            
            models_data = response.json()
            return [
                Model(
                    id=model["id"],
                    owned_by=model.get("owned_by", "openai"),
                    context_window=None,  # Audio models don't have context windows
                )
                for model in models_data["data"]
                if model["id"].startswith("whisper")
            ]
        except Exception:
            # Handle the case when the API key is not valid for model listing
            return []

    def _get_api_kwargs(
        self, language: Optional[str] = None, prompt: Optional[str] = None
    ) -> Dict[str, Any]:
        """Get kwargs for API calls."""
        kwargs = {
            "model": self.get_model_name(),
        }

        if language:
            kwargs["language"] = language
        if prompt:
            kwargs["prompt"] = prompt

        return kwargs

    def transcribe(
        self,
        audio_file: Union[str, BinaryIO],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
    ) -> TranscriptionResponse:
        """Transcribe audio to text using OpenAI's Whisper model."""
        kwargs = self._get_api_kwargs(language, prompt)

        # Handle file input
        if isinstance(audio_file, str):
            # For file path, open and send as multipart form data
            with open(audio_file, "rb") as f:
                files = {"file": (audio_file, f, "audio/mpeg")}
                response = self.client.post(
                    f"{self.base_url}/audio/transcriptions",
                    headers=self._get_headers(),
                    files=files,
                    data=kwargs
                )
        else:
            # For BinaryIO, send the file object directly
            filename = getattr(audio_file, 'name', 'audio.mp3')
            files = {"file": (filename, audio_file, "audio/mpeg")}
            response = self.client.post(
                f"{self.base_url}/audio/transcriptions",
                headers=self._get_headers(),
                files=files,
                data=kwargs
            )

        self._handle_error(response)
        response_data = response.json()

        return TranscriptionResponse(
            text=response_data["text"],
            language=language,  # OpenAI doesn't return detected language
            model=self.get_model_name(),
            provider=self.provider,
        )

    async def atranscribe(
        self,
        audio_file: Union[str, BinaryIO],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
    ) -> TranscriptionResponse:
        """Async transcribe audio to text using OpenAI's Whisper model."""
        kwargs = self._get_api_kwargs(language, prompt)

        # Handle file input
        if isinstance(audio_file, str):
            # For file path, open and send as multipart form data
            with open(audio_file, "rb") as f:
                files = {"file": (audio_file, f, "audio/mpeg")}
                response = await self.async_client.post(
                    f"{self.base_url}/audio/transcriptions",
                    headers=self._get_headers(),
                    files=files,
                    data=kwargs
                )
        else:
            # For BinaryIO, send the file object directly
            filename = getattr(audio_file, 'name', 'audio.mp3')
            files = {"file": (filename, audio_file, "audio/mpeg")}
            response = await self.async_client.post(
                f"{self.base_url}/audio/transcriptions",
                headers=self._get_headers(),
                files=files,
                data=kwargs
            )

        self._handle_error(response)
        response_data = response.json()

        return TranscriptionResponse(
            text=response_data["text"],
            language=language,  # OpenAI doesn't return detected language
            model=self.get_model_name(),
            provider=self.provider,
        )



"""Integration tests for embedding providers."""

from unittest.mock import AsyncMock, MagicMock

import pytest

from esperanto.providers.embedding.azure import AzureEmbeddingModel
from esperanto.providers.embedding.google import GoogleEmbeddingModel

# @pytest.fixture
# def transformers_model():
#     """Create a transformers model instance."""
#     return AIFactory.create_embedding(
#         provider="transformers",
#         model_name="bert-base-uncased",
#         config={
#             "device": "cpu",
#             "pooling_strategy": "mean",
#         },
#     )


# def test_transformers_embedding(transformers_model):
#     """Test transformers embedding generation."""
#     texts = [
#         "This is a test sentence.",
#         "Another example for embedding.",
#     ]

#     embeddings = transformers_model.embed(texts)

#     # Basic validation
#     assert len(embeddings) == len(texts)
#     assert all(isinstance(emb, list) for emb in embeddings)
#     assert all(isinstance(val, float) for emb in embeddings for val in emb)
#     # BERT base has 768-dimensional embeddings
#     assert all(len(emb) == 768 for emb in embeddings)

#     # Verify embeddings are different for different texts
#     assert embeddings[0] != embeddings[1]


# @pytest.mark.asyncio
# async def test_transformers_async_embedding(transformers_model):
#     """Test async transformers embedding generation."""
#     texts = [
#         "This is a test sentence.",
#         "Another example for embedding.",
#     ]

#     embeddings = await transformers_model.aembed(texts)

#     # Basic validation
#     assert len(embeddings) == len(texts)
#     assert all(isinstance(emb, list) for emb in embeddings)
#     assert all(isinstance(val, float) for emb in embeddings for val in emb)
#     # BERT base has 768-dimensional embeddings
#     assert all(len(emb) == 768 for emb in embeddings)

#     # Verify embeddings are different for different texts
#     assert embeddings[0] != embeddings[1]


# Note, if the mock code is commented out, and the API Key and Base URL replaced with actual
# values then an actual model can be called. I assumed that was the purpose of the tests in
# this file.
@pytest.fixture
def azure_model():
    """Create a Azure embedding model instance with mocked responses."""
    model = AzureEmbeddingModel(api_key="AZURE_OPENAI_API_KEY",
                                base_url="AZURE_OPENAI_ENDPOINT",
                                api_version="2024-02-01",
                                model_name="text-embedding-3-large")

    # Mock embed and aembed methods w/ 256-dimensional embedding
    sample_embedding = [[0.1, 0.2, 0.3, 0.4] * 64,
                        [0.4, 0.3, 0.2, 0.1] * 64]
    model.embed = MagicMock(return_value=sample_embedding)
    model.aembed = AsyncMock(return_value=sample_embedding)

    return model

@pytest.mark.asyncio
async def test_azure_embedding(azure_model):
    """Test Azure embedding generation."""
    texts = [
        "This is a test sentence.",
        "Another example for embedding.",
    ]

    # text-embedding-3-large defaults to 3072-dimensional embedding
    # text-embedding-3-small defaults to 1536-dimensional embedding
    # mock returns a 256-dimensional embedding
    dimensions = 256

    embeddings = azure_model.embed(texts, dimensions=dimensions)
    aembeddings = await azure_model.aembed(texts, dimensions=dimensions)

    assert embeddings == aembeddings

    # Basic validation
    assert isinstance(embeddings, list)
    assert all(isinstance(emb, list) for emb in embeddings)
    assert all(isinstance(val, float) for emb in embeddings for val in emb)

    # Data check - our mock returns a 256-dim embedding
    assert all(len(emb) == dimensions for emb in embeddings)
    assert embeddings[0] != embeddings[1]


@pytest.fixture
def google_model():
    """Create a Google embedding model instance with mocked responses."""
    model = GoogleEmbeddingModel(api_key="test-key")

    # Mock embed and aembed methods
    sample_embedding = [[0.1, 0.2, 0.3, 0.4] * 64]  # 256-dimensional embedding
    model.embed = MagicMock(return_value=sample_embedding)
    model.aembed = AsyncMock(return_value=sample_embedding)

    return model


def test_google_embedding(google_model):
    """Test Google embedding generation."""
    texts = [
        "This is a test sentence.",
        "Another example for embedding.",
    ]

    embeddings = google_model.embed(texts)

    # Verify the method was called with the correct texts
    google_model.embed.assert_called_once_with(texts)

    # Basic validation
    assert isinstance(embeddings, list)
    assert all(isinstance(emb, list) for emb in embeddings)
    assert all(isinstance(val, float) for emb in embeddings for val in emb)

    # Length check - our mock returns a 256-dim embedding
    assert len(embeddings[0]) == 256


@pytest.mark.asyncio
async def test_google_async_embedding(google_model):
    """Test async Google embedding generation."""
    texts = [
        "This is a test sentence.",
        "Another example for embedding.",
    ]

    embeddings = await google_model.aembed(texts)

    # Verify the method was called with the correct texts
    google_model.aembed.assert_called_once_with(texts)

    # Basic validation
    assert isinstance(embeddings, list)
    assert all(isinstance(emb, list) for emb in embeddings)
    assert all(isinstance(val, float) for emb in embeddings for val in emb)

    # Length check - our mock returns a 256-dim embedding
    assert len(embeddings[0]) == 256


# def test_transformers_pooling_strategies(transformers_model):
#     """Test different pooling strategies."""
#     text = ["Test sentence for pooling."]

#     # Test mean pooling (default)
#     mean_embeddings = transformers_model.embed(text)

#     # Test max pooling
#     transformers_model.pooling_config.strategy = "max"
#     max_embeddings = transformers_model.embed(text)

#     # Test cls pooling
#     transformers_model.pooling_config.strategy = "cls"
#     cls_embeddings = transformers_model.embed(text)

#     # All should have same dimensions but different values
#     assert len(mean_embeddings[0]) == len(max_embeddings[0]) == len(cls_embeddings[0])
#     assert mean_embeddings != max_embeddings
#     assert mean_embeddings != cls_embeddings
#     assert max_embeddings != cls_embeddings


# def test_transformers_batch_processing(transformers_model):
#     """Test batch processing of texts."""
#     # Create a list of texts longer than the default batch size
#     texts = [
#         f"Test sentence number {i}." for i in range(40)
#     ]  # Default batch size is 32

#     embeddings = transformers_model.embed(texts)

#     # Verify all texts were processed
#     assert len(embeddings) == len(texts)
#     assert all(len(emb) == 768 for emb in embeddings)


# @pytest.mark.asyncio
# async def test_transformers_async_batch_processing(transformers_model):
#     """Test async batch processing of texts."""
#     # Create a list of texts longer than the default batch size
#     texts = [
#         f"Test sentence number {i}." for i in range(40)
#     ]  # Default batch size is 32

#     embeddings = await transformers_model.aembed(texts)

#     # Verify all texts were processed
#     assert len(embeddings) == len(texts)
#     assert all(len(emb) == 768 for emb in embeddings)


# def test_transformers_error_handling(transformers_model):
#     """Test error handling for invalid inputs."""
#     # Test empty input
#     with pytest.raises(ValueError, match="Texts cannot be empty"):
#         transformers_model.embed([])

#     # Test None input
#     with pytest.raises(ValueError, match="Texts cannot be empty"):
#         transformers_model.embed(None)


# @pytest.mark.asyncio
# async def test_transformers_async_error_handling(transformers_model):
#     """Test async error handling for invalid inputs."""
#     # Test empty input
#     with pytest.raises(ValueError, match="Texts cannot be empty"):
#         await transformers_model.aembed([])

#     # Test None input
#     with pytest.raises(ValueError, match="Texts cannot be empty"):
#         await transformers_model.aembed(None)


# def test_transformers_device_config():
#     """Test device configuration."""
#     # Test explicit CPU config
#     cpu_model = AIFactory.create_embedding(
#         provider="transformers",
#         model_name="bert-base-uncased",
#         config={"device": "cpu"},
#     )
#     assert cpu_model.device == "cpu"

#     # Test auto device config with CPU (no CUDA or MPS)
#     with (
#         patch("torch.cuda.is_available", return_value=False),
#         patch("torch.backends.mps.is_available", return_value=False),
#     ):
#         cpu_auto_model = AIFactory.create_embedding(
#             provider="transformers",
#             model_name="bert-base-uncased",
#             config={"device": "auto"},
#         )
#         assert cpu_auto_model.device == "cpu"

#     # Test auto device config with MPS (Apple Silicon)
#     with (
#         patch("torch.cuda.is_available", return_value=False),
#         patch("torch.backends.mps.is_available", return_value=True),
#     ):
#         mps_auto_model = AIFactory.create_embedding(
#             provider="transformers",
#             model_name="bert-base-uncased",
#             config={"device": "auto"},
#         )
#         assert mps_auto_model.device == "mps"



"""Tests for the AIFactory class."""

import pytest
from unittest.mock import patch, MagicMock

from esperanto.factory import AIFactory


# Note: Caching functionality has been removed from AIFactory.
# Tests now verify that factory creates new instances each time.


"""Tests for the OpenAI-compatible STT provider."""
import pytest
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, Mock
from io import BytesIO

from esperanto.providers.stt.openai_compatible import OpenAICompatibleSpeechToTextModel


@pytest.fixture
def mock_stt_response():
    """Mock STT response data."""
    return {
        "text": "Hello, this is a test transcription from OpenAI-compatible endpoint."
    }


@pytest.fixture
def mock_openai_compatible_models_response():
    """Mock HTTP response for OpenAI-compatible models API."""
    return {
        "object": "list",
        "data": [
            {
                "id": "faster-whisper-large-v3",
                "object": "model",
                "owned_by": "custom"
            },
            {
                "id": "whisper-1",
                "object": "model",
                "owned_by": "custom"
            }
        ]
    }


@pytest.fixture
def mock_httpx_clients(mock_stt_response, mock_openai_compatible_models_response):
    """Mock httpx clients for OpenAI-compatible STT."""
    client = Mock()
    async_client = AsyncMock()

    # Mock HTTP response objects
    def make_response(status_code, json_data=None):
        response = Mock()
        response.status_code = status_code
        if json_data is not None:
            response.json.return_value = json_data
        return response

    def make_async_response(status_code, json_data=None):
        response = AsyncMock()
        response.status_code = status_code
        if json_data is not None:
            response.json = Mock(return_value=json_data)
        return response

    # Configure responses based on URL
    def mock_post_side_effect(url, **kwargs):
        if url.endswith("/audio/transcriptions"):
            return make_response(200, json_data=mock_stt_response)
        return make_response(404, json_data={"error": "Not found"})

    def mock_get_side_effect(url, **kwargs):
        if url.endswith("/models"):
            return make_response(200, json_data=mock_openai_compatible_models_response)
        return make_response(404, json_data={"error": "Not found"})

    async def mock_async_post_side_effect(url, **kwargs):
        if url.endswith("/audio/transcriptions"):
            return make_async_response(200, json_data=mock_stt_response)
        return make_async_response(404, json_data={"error": "Not found"})

    async def mock_async_get_side_effect(url, **kwargs):
        if url.endswith("/models"):
            return make_async_response(200, json_data=mock_openai_compatible_models_response)
        return make_async_response(404, json_data={"error": "Not found"})

    # Mock synchronous HTTP client
    client.post.side_effect = mock_post_side_effect
    client.get.side_effect = mock_get_side_effect

    # Mock async HTTP client
    async_client.post.side_effect = mock_async_post_side_effect
    async_client.get.side_effect = mock_async_get_side_effect

    return client, async_client


@pytest.fixture
def stt_model(mock_httpx_clients):
    """Create a STT model instance with mocked HTTP clients."""
    model = OpenAICompatibleSpeechToTextModel(
        model_name="faster-whisper-large-v3",
        api_key="test-key",
        base_url="http://localhost:8000"
    )
    model.client, model.async_client = mock_httpx_clients
    return model


def test_init_with_config():
    """Test model initialization with config."""
    model = OpenAICompatibleSpeechToTextModel(
        model_name="faster-whisper-large-v3",
        api_key="test-key",
        base_url="http://localhost:8000"
    )
    assert model.model_name == "faster-whisper-large-v3"
    assert model.provider == "openai-compatible"
    assert model.base_url == "http://localhost:8000"
    assert model.api_key == "test-key"


def test_init_with_env_vars(monkeypatch):
    """Test model initialization with environment variables."""
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL", "http://localhost:9000")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY", "env-key")

    model = OpenAICompatibleSpeechToTextModel(model_name="test-model")
    assert model.base_url == "http://localhost:9000"
    assert model.api_key == "env-key"


def test_init_missing_base_url(monkeypatch):
    """Test that initialization fails without base URL."""
    # Clear all environment variables
    monkeypatch.delenv("OPENAI_COMPATIBLE_BASE_URL_STT", raising=False)
    monkeypatch.delenv("OPENAI_COMPATIBLE_BASE_URL", raising=False)

    with pytest.raises(ValueError, match="OpenAI-compatible base URL is required"):
        OpenAICompatibleSpeechToTextModel(model_name="test-model")


def test_base_url_trailing_slash():
    """Test that trailing slash is stripped from base URL."""
    model = OpenAICompatibleSpeechToTextModel(
        model_name="test-model",
        base_url="http://localhost:8000/"
    )
    assert model.base_url == "http://localhost:8000"


def test_timeout_configuration():
    """Test timeout configuration."""
    model = OpenAICompatibleSpeechToTextModel(
        model_name="test-model",
        base_url="http://localhost:8000",
        config={"timeout": 600}
    )
    assert model.timeout == 600


def test_transcribe_with_file_path(stt_model, tmp_path):
    """Test synchronous transcription with file path."""
    # Create a temporary audio file
    audio_file = tmp_path / "test.mp3"
    audio_file.write_bytes(b"fake audio data")

    response = stt_model.transcribe(str(audio_file))

    # Verify HTTP POST was called
    stt_model.client.post.assert_called_once()
    call_args = stt_model.client.post.call_args

    # Check URL
    assert call_args[0][0] == "http://localhost:8000/audio/transcriptions"

    # Check headers
    headers = call_args[1]["headers"]
    assert "Authorization" in headers
    assert headers["Authorization"] == "Bearer test-key"

    # Check multipart form data
    files = call_args[1]["files"]
    assert "file" in files

    # Check form data
    data = call_args[1]["data"]
    assert data["model"] == "faster-whisper-large-v3"

    # Check response
    assert response.text == "Hello, this is a test transcription from OpenAI-compatible endpoint."
    assert response.model == "faster-whisper-large-v3"


def test_transcribe_with_file_object(stt_model):
    """Test synchronous transcription with file object."""
    audio_data = BytesIO(b"fake audio data")
    audio_data.name = "test.mp3"

    response = stt_model.transcribe(audio_data)

    # Verify HTTP POST was called
    stt_model.client.post.assert_called_once()
    call_args = stt_model.client.post.call_args

    # Check URL
    assert call_args[0][0] == "http://localhost:8000/audio/transcriptions"

    # Check multipart form data
    files = call_args[1]["files"]
    assert "file" in files

    # Check response
    assert response.text == "Hello, this is a test transcription from OpenAI-compatible endpoint."
    assert response.model == "faster-whisper-large-v3"


@pytest.mark.asyncio
async def test_atranscribe_with_file_path(stt_model, tmp_path):
    """Test asynchronous transcription with file path."""
    # Create a temporary audio file
    audio_file = tmp_path / "test.wav"
    audio_file.write_bytes(b"fake audio data")

    response = await stt_model.atranscribe(str(audio_file))

    # Verify async HTTP POST was called
    stt_model.async_client.post.assert_called_once()
    call_args = stt_model.async_client.post.call_args

    # Check URL
    assert call_args[0][0] == "http://localhost:8000/audio/transcriptions"

    # Check headers
    headers = call_args[1]["headers"]
    assert "Authorization" in headers
    assert headers["Authorization"] == "Bearer test-key"

    # Check response
    assert response.text == "Hello, this is a test transcription from OpenAI-compatible endpoint."
    assert response.model == "faster-whisper-large-v3"


@pytest.mark.asyncio
async def test_atranscribe_with_file_object(stt_model):
    """Test asynchronous transcription with file object."""
    audio_data = BytesIO(b"fake audio data")
    audio_data.name = "test.wav"

    response = await stt_model.atranscribe(audio_data)

    # Verify async HTTP POST was called
    stt_model.async_client.post.assert_called_once()

    # Check response
    assert response.text == "Hello, this is a test transcription from OpenAI-compatible endpoint."


def test_transcribe_with_language_and_prompt(stt_model, tmp_path):
    """Test transcription with language and prompt parameters."""
    audio_file = tmp_path / "test.mp3"
    audio_file.write_bytes(b"fake audio data")

    response = stt_model.transcribe(
        str(audio_file),
        language="en",
        prompt="This is a technical discussion"
    )

    # Check that language and prompt were passed in form data
    call_args = stt_model.client.post.call_args
    data = call_args[1]["data"]
    assert data["language"] == "en"
    assert data["prompt"] == "This is a technical discussion"


def test_models(stt_model):
    """Test that the models property works with HTTP."""
    models = stt_model.models

    # Verify HTTP GET was called
    stt_model.client.get.assert_called_with(
        "http://localhost:8000/models",
        headers={"Authorization": "Bearer test-key"}
    )

    # Check that models are returned
    assert len(models) == 2
    assert models[0].id == "faster-whisper-large-v3"
    assert models[1].id == "whisper-1"
    # Model type is None when not explicitly provided by the API
    assert models[0].type is None
    assert models[1].type is None
    assert models[0].owned_by == "custom"
    assert models[1].owned_by == "custom"


def test_models_fallback():
    """Test that models property falls back gracefully when endpoint doesn't support it."""
    model = OpenAICompatibleSpeechToTextModel(
        api_key="test-key",
        model_name="test-model",
        base_url="http://localhost:8000"
    )

    # Mock client that throws exception for models endpoint
    client = Mock()
    client.get.side_effect = Exception("Connection error")
    model.client = client

    models = model.models

    # Should return empty list
    assert models == []


def test_error_handling():
    """Test error handling for HTTP errors."""
    model = OpenAICompatibleSpeechToTextModel(
        api_key="test-key",
        model_name="test-model",
        base_url="http://localhost:8000"
    )

    # Mock client that returns error response
    client = Mock()
    response = Mock()
    response.status_code = 500
    response.text = "Internal server error"
    response.json.side_effect = Exception("Not JSON")
    client.post.return_value = response
    model.client = client

    audio_data = BytesIO(b"fake audio data")
    with pytest.raises(RuntimeError, match="OpenAI-compatible STT endpoint error"):
        model.transcribe(audio_data)


def test_get_default_model():
    """Test default model name."""
    model = OpenAICompatibleSpeechToTextModel(
        base_url="http://localhost:8000"
    )
    assert model._get_default_model() == "whisper-1"


def test_provider_name(stt_model):
    """Test provider name."""
    assert stt_model.provider == "openai-compatible"


def test_get_model_name(stt_model):
    """Test getting model name."""
    assert stt_model.get_model_name() == "faster-whisper-large-v3"


def test_get_headers(stt_model):
    """Test header generation."""
    headers = stt_model._get_headers()
    assert headers["Authorization"] == "Bearer test-key"


def test_get_api_kwargs(stt_model):
    """Test API kwargs generation."""
    kwargs = stt_model._get_api_kwargs("en", "technical discussion")
    assert kwargs["model"] == "faster-whisper-large-v3"
    assert kwargs["language"] == "en"
    assert kwargs["prompt"] == "technical discussion"


def test_provider_specific_env_var_precedence(monkeypatch):
    """Test that provider-specific env vars take precedence over generic ones."""
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL_STT", "http://stt-specific:1234")
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL", "http://generic:5678")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY_STT", "stt-specific-key")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY", "generic-key")

    model = OpenAICompatibleSpeechToTextModel(model_name="test-model")
    assert model.base_url == "http://stt-specific:1234"
    assert model.api_key == "stt-specific-key"


def test_fallback_to_generic_env_var(monkeypatch):
    """Test fallback to generic env vars when provider-specific ones are not set."""
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL", "http://generic:5678")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY", "generic-key")

    model = OpenAICompatibleSpeechToTextModel(model_name="test-model")
    assert model.base_url == "http://generic:5678"
    assert model.api_key == "generic-key"


def test_config_overrides_provider_specific_env_vars(monkeypatch):
    """Test that config parameters override provider-specific env vars."""
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL_STT", "http://stt-env:1234")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY_STT", "stt-env-key")

    model = OpenAICompatibleSpeechToTextModel(
        model_name="test-model",
        base_url="http://config:9090",
        api_key="config-key"
    )
    assert model.base_url == "http://config:9090"
    assert model.api_key == "config-key"


def test_direct_params_override_all_env_vars(monkeypatch):
    """Test that direct parameters override all environment variables."""
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL_STT", "http://stt-env:1234")
    monkeypatch.setenv("OPENAI_COMPATIBLE_BASE_URL", "http://generic-env:5678")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY_STT", "stt-env-key")
    monkeypatch.setenv("OPENAI_COMPATIBLE_API_KEY", "generic-env-key")

    model = OpenAICompatibleSpeechToTextModel(
        model_name="test-model",
        base_url="http://direct:3000",
        api_key="direct-key"
    )
    assert model.base_url == "http://direct:3000"
    assert model.api_key == "direct-key"


def test_error_message_mentions_both_env_vars(monkeypatch):
    """Test that error message mentions both provider-specific and generic env vars."""
    # Clear all environment variables
    monkeypatch.delenv("OPENAI_COMPATIBLE_BASE_URL_STT", raising=False)
    monkeypatch.delenv("OPENAI_COMPATIBLE_BASE_URL", raising=False)

    with pytest.raises(ValueError) as exc_info:
        OpenAICompatibleSpeechToTextModel(model_name="test-model", api_key="test-key")

    error_message = str(exc_info.value)
    assert "OPENAI_COMPATIBLE_BASE_URL_STT" in error_message
    assert "OPENAI_COMPATIBLE_BASE_URL" in error_message


"""Tests for Message.thinking and Message.cleaned_content properties."""

import pytest

from esperanto.common_types import Message


class TestMessageThinkingProperty:
    """Tests for the thinking property."""

    def test_thinking_extracts_single_block(self):
        """Test extracting content from a single think block."""
        content = "<think>Let me think about this.</think>\n\nThe answer is 42."
        msg = Message(content=content, role="assistant")
        assert msg.thinking == "Let me think about this."

    def test_thinking_extracts_multiline_block(self):
        """Test extracting multiline content from think block."""
        content = """<think>
First, I need to consider X.
Then, I should look at Y.
Finally, Z is important.
</think>

The answer is 42."""
        msg = Message(content=content, role="assistant")
        assert "First, I need to consider X." in msg.thinking
        assert "Then, I should look at Y." in msg.thinking
        assert "Finally, Z is important." in msg.thinking

    def test_thinking_concatenates_multiple_blocks(self):
        """Test that multiple think blocks are concatenated."""
        content = """<think>First thought</think>

Some response

<think>Second thought</think>

More response"""
        msg = Message(content=content, role="assistant")
        assert msg.thinking == "First thought\n\nSecond thought"

    def test_thinking_returns_none_without_tags(self):
        """Test that thinking returns None when no think tags present."""
        content = '{"name": "John", "age": 30}'
        msg = Message(content=content, role="assistant")
        assert msg.thinking is None

    def test_thinking_returns_none_for_empty_content(self):
        """Test that thinking returns None for None content."""
        msg = Message(content=None, role="assistant")
        assert msg.thinking is None

    def test_thinking_returns_none_for_empty_think_tags(self):
        """Test that thinking returns None when think tags are empty."""
        content = "<think>\n\n</think>\n\n{\"result\": 42}"
        msg = Message(content=content, role="assistant")
        assert msg.thinking is None

    def test_thinking_strips_whitespace(self):
        """Test that thinking content is stripped of leading/trailing whitespace."""
        content = "<think>   \n  Padded content  \n   </think>"
        msg = Message(content=content, role="assistant")
        assert msg.thinking == "Padded content"


class TestMessageCleanedContentProperty:
    """Tests for the cleaned_content property."""

    def test_cleaned_content_removes_think_block(self):
        """Test that cleaned_content removes think blocks."""
        content = "<think>Let me think.</think>\n\n{\"answer\": 42}"
        msg = Message(content=content, role="assistant")
        assert msg.cleaned_content == '{"answer": 42}'

    def test_cleaned_content_removes_multiple_blocks(self):
        """Test that cleaned_content removes multiple think blocks."""
        content = """<think>First thought</think>

Some response

<think>Second thought</think>

More response"""
        msg = Message(content=content, role="assistant")
        cleaned = msg.cleaned_content
        assert "<think>" not in cleaned
        assert "</think>" not in cleaned
        assert "First thought" not in cleaned
        assert "Second thought" not in cleaned
        assert "Some response" in cleaned
        assert "More response" in cleaned

    def test_cleaned_content_returns_full_content_without_tags(self):
        """Test that cleaned_content returns full content when no think tags."""
        content = '{"name": "John", "age": 30}'
        msg = Message(content=content, role="assistant")
        assert msg.cleaned_content == content

    def test_cleaned_content_returns_empty_string_for_none(self):
        """Test that cleaned_content returns empty string for None content."""
        msg = Message(content=None, role="assistant")
        assert msg.cleaned_content == ""

    def test_cleaned_content_handles_empty_think_tags(self):
        """Test that cleaned_content handles empty think tags correctly."""
        content = "<think>\n\n</think>\n\n{\"result\": 42}"
        msg = Message(content=content, role="assistant")
        assert msg.cleaned_content == '{"result": 42}'

    def test_cleaned_content_cleans_excessive_newlines(self):
        """Test that cleaned_content normalizes excessive newlines."""
        content = "<think>Thinking...</think>\n\n\n\n\nThe result"
        msg = Message(content=content, role="assistant")
        # Should not have more than 2 consecutive newlines
        assert "\n\n\n" not in msg.cleaned_content
        assert "The result" in msg.cleaned_content


class TestMessageThinkingIntegration:
    """Integration tests for thinking/cleaned_content with real-world examples."""

    def test_qwen_style_response(self):
        """Test parsing Qwen3-style response with think tags."""
        content = """<think>
Okay, the user wants a JSON with name and age for a fictional person.
Let me create something reasonable.
Name: Elena Voss
Age: 34
</think>

{"name": "Elena Voss", "age": 34}"""
        msg = Message(content=content, role="assistant")

        assert msg.thinking is not None
        assert "Elena Voss" in msg.thinking
        assert "Age: 34" in msg.thinking

        assert msg.cleaned_content == '{"name": "Elena Voss", "age": 34}'
        assert "<think>" not in msg.cleaned_content

    def test_empty_string_content(self):
        """Test with empty string content."""
        msg = Message(content="", role="assistant")
        assert msg.thinking is None
        assert msg.cleaned_content == ""

    def test_original_content_unchanged(self):
        """Test that original content property is unchanged."""
        content = "<think>Reasoning</think>\n\nResult"
        msg = Message(content=content, role="assistant")

        # Original content should be preserved
        assert msg.content == content
        # Properties should parse it differently
        assert msg.thinking == "Reasoning"
        assert msg.cleaned_content == "Result"



#!/usr/bin/env python3
"""Provider comparison example for tool calling with Esperanto.

This example demonstrates Esperanto's key value proposition:
the same tool calling code works across all providers.

Environment variables (set the ones for providers you want to test):
- OPENAI_API_KEY: OpenAI API key
- ANTHROPIC_API_KEY: Anthropic API key
- GOOGLE_API_KEY: Google (Gemini) API key
- GROQ_API_KEY: Groq API key
- MISTRAL_API_KEY: Mistral API key
- DEEPSEEK_API_KEY: DeepSeek API key
- XAI_API_KEY: xAI API key

Run with:
    python examples/tool_calling/provider_comparison.py
"""

import json
import os
from typing import Any, Dict, List, Optional, Tuple

from esperanto import AIFactory
from esperanto.common_types import Tool, ToolFunction


def create_tools() -> List[Tool]:
    """Create a set of tools that work across all providers."""

    # Simple calculator tool
    calculator = Tool(
        type="function",
        function=ToolFunction(
            name="calculate",
            description="Perform a mathematical calculation",
            parameters={
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "Mathematical expression to evaluate",
                    }
                },
                "required": ["expression"],
            },
        ),
    )

    # Weather tool
    weather = Tool(
        type="function",
        function=ToolFunction(
            name="get_weather",
            description="Get the current weather for a city",
            parameters={
                "type": "object",
                "properties": {
                    "city": {
                        "type": "string",
                        "description": "City name",
                    }
                },
                "required": ["city"],
            },
        ),
    )

    return [calculator, weather]


def get_available_providers() -> List[Tuple[str, str, str]]:
    """Get list of available providers based on environment variables.

    Returns:
        List of (provider_name, model_name, env_var_name) tuples.
    """
    providers = [
        ("openai", "gpt-4o-mini", "OPENAI_API_KEY"),
        ("anthropic", "claude-3-5-haiku-20241022", "ANTHROPIC_API_KEY"),
        ("google", "gemini-2.0-flash", "GOOGLE_API_KEY"),
        ("groq", "llama-3.3-70b-versatile", "GROQ_API_KEY"),
        ("mistral", "mistral-small-latest", "MISTRAL_API_KEY"),
        ("deepseek", "deepseek-chat", "DEEPSEEK_API_KEY"),
        ("xai", "grok-2-latest", "XAI_API_KEY"),
    ]

    available = []
    for provider, model, env_var in providers:
        if os.getenv(env_var):
            available.append((provider, model, env_var))

    return available


def test_provider(
    provider: str, model: str, tools: List[Tool], query: str
) -> Dict[str, Any]:
    """Test tool calling with a specific provider.

    Returns:
        Dict with results including success status, tool calls, and any errors.
    """
    result: Dict[str, Any] = {
        "provider": provider,
        "model": model,
        "success": False,
        "tool_calls": None,
        "content": None,
        "error": None,
    }

    try:
        # Create the model - same code for all providers!
        llm = AIFactory.create_language(provider=provider, model_name=model)

        # Make the request - identical across providers
        messages = [{"role": "user", "content": query}]
        response = llm.chat_complete(messages, tools=tools)

        message = response.choices[0].message
        result["success"] = True

        if message.tool_calls:
            result["tool_calls"] = [
                {
                    "name": tc.function.name,
                    "arguments": json.loads(tc.function.arguments),
                }
                for tc in message.tool_calls
            ]
        else:
            result["content"] = message.content

    except Exception as e:
        result["error"] = str(e)

    return result


def format_result(result: Dict[str, Any]) -> str:
    """Format a test result for display."""
    lines = [f"  Provider: {result['provider']} ({result['model']})"]

    if result["error"]:
        lines.append(f"  Status: FAILED")
        lines.append(f"  Error: {result['error'][:100]}...")
    elif result["tool_calls"]:
        lines.append(f"  Status: SUCCESS (tool call)")
        for tc in result["tool_calls"]:
            lines.append(f"  Tool: {tc['name']}")
            lines.append(f"  Args: {tc['arguments']}")
    else:
        lines.append(f"  Status: SUCCESS (text response)")
        content = result["content"] or ""
        lines.append(f"  Response: {content[:100]}...")

    return "\n".join(lines)


def main():
    print("=== Provider Comparison: Tool Calling ===\n")

    # Get available providers
    available = get_available_providers()

    if not available:
        print("No provider API keys found!")
        print("\nPlease set at least one of these environment variables:")
        print("  - OPENAI_API_KEY")
        print("  - ANTHROPIC_API_KEY")
        print("  - GOOGLE_API_KEY")
        print("  - GROQ_API_KEY")
        print("  - MISTRAL_API_KEY")
        print("  - DEEPSEEK_API_KEY")
        print("  - XAI_API_KEY")
        return

    print(f"Found {len(available)} provider(s) with API keys:\n")
    for provider, model, _ in available:
        print(f"  - {provider}: {model}")
    print()

    # Create tools - same tools work for all providers
    tools = create_tools()
    print(f"Tools defined: {[t.function.name for t in tools]}\n")

    # Test queries
    test_cases = [
        ("What is 15 * 23?", "Should trigger calculate tool"),
        ("What's the weather in Paris?", "Should trigger get_weather tool"),
    ]

    for query, description in test_cases:
        print("=" * 60)
        print(f"Query: {query}")
        print(f"Expected: {description}")
        print("=" * 60)
        print()

        for provider, model, _ in available:
            result = test_provider(provider, model, tools, query)
            print(format_result(result))
            print()

    # Summary
    print("=" * 60)
    print("KEY TAKEAWAY")
    print("=" * 60)
    print("""
The same tool definition and calling code works across all providers:

    # Define tools once
    tools = [Tool(...), Tool(...)]

    # Works with ANY provider
    model = AIFactory.create_language(provider="openai", ...)
    model = AIFactory.create_language(provider="anthropic", ...)
    model = AIFactory.create_language(provider="google", ...)

    # Identical API call
    response = model.chat_complete(messages, tools=tools)

Esperanto handles all the format conversion internally:
- OpenAI format: {"type": "function", "function": {...}}
- Anthropic format: {"name": ..., "input_schema": {...}}
- Google format: {"function_declarations": [...]}

You write code once, it works everywhere.
""")

    print("=== Example Complete ===")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Multiple tools example with Esperanto.

This example demonstrates using multiple tools:
1. Define several tools for different purposes
2. Let the model choose which tool(s) to call
3. Handle different tool call types

Environment variables required:
- OPENAI_API_KEY: Your OpenAI API key

Run with:
    python examples/tool_calling/multiple_tools.py
"""

import json
import os

from esperanto import AIFactory
from esperanto.common_types import Tool, ToolFunction


def create_tools():
    """Create a set of tools for a hypothetical assistant."""

    # Weather tool
    weather_tool = Tool(
        type="function",
        function=ToolFunction(
            name="get_weather",
            description="Get current weather for a location",
            parameters={
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "City name, e.g., 'Paris, France'",
                    }
                },
                "required": ["location"],
            },
        ),
    )

    # Calculator tool
    calculator_tool = Tool(
        type="function",
        function=ToolFunction(
            name="calculate",
            description="Perform mathematical calculations",
            parameters={
                "type": "object",
                "properties": {
                    "expression": {
                        "type": "string",
                        "description": "Mathematical expression to evaluate, e.g., '2 + 2' or '15 * 7'",
                    }
                },
                "required": ["expression"],
            },
        ),
    )

    # Search tool
    search_tool = Tool(
        type="function",
        function=ToolFunction(
            name="search_web",
            description="Search the web for information",
            parameters={
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "Search query",
                    },
                    "num_results": {
                        "type": "integer",
                        "description": "Number of results to return (1-10)",
                        "minimum": 1,
                        "maximum": 10,
                    },
                },
                "required": ["query"],
            },
        ),
    )

    # Time tool
    time_tool = Tool(
        type="function",
        function=ToolFunction(
            name="get_current_time",
            description="Get the current time in a specific timezone",
            parameters={
                "type": "object",
                "properties": {
                    "timezone": {
                        "type": "string",
                        "description": "Timezone name, e.g., 'America/New_York' or 'Europe/London'",
                    }
                },
                "required": ["timezone"],
            },
        ),
    )

    return [weather_tool, calculator_tool, search_tool, time_tool]


def main():
    # Check for API key
    if not os.getenv("OPENAI_API_KEY"):
        print("Please set OPENAI_API_KEY environment variable")
        print("Export it with: export OPENAI_API_KEY='your-key-here'")
        return

    # Create tools
    tools = create_tools()

    # Create model
    model = AIFactory.create_language(
        provider="openai",
        model_name="gpt-4o-mini",
        config={"tools": tools},
    )

    print("=== Multiple Tools Example ===\n")
    print(f"Available tools: {[t.function.name for t in tools]}\n")

    # Test different queries that should trigger different tools
    test_queries = [
        "What's 15 multiplied by 23?",
        "What's the weather in Berlin?",
        "What time is it in Tokyo?",
        "Search for the latest news about AI",
    ]

    for query in test_queries:
        print(f"User: {query}")
        messages = [{"role": "user", "content": query}]

        response = model.chat_complete(messages)
        message = response.choices[0].message

        if message.tool_calls:
            for tool_call in message.tool_calls:
                args = json.loads(tool_call.function.arguments)
                print(f"  -> Tool: {tool_call.function.name}")
                print(f"     Args: {args}")
        else:
            print(f"  -> Response: {message.content[:100]}...")

        print()

    # Test a query that might trigger multiple tools (parallel tool calls)
    print("--- Testing parallel tool calls ---\n")
    complex_query = "What's the weather in New York and what time is it there?"
    print(f"User: {complex_query}")

    messages = [{"role": "user", "content": complex_query}]

    # Enable parallel tool calls
    response = model.chat_complete(messages, parallel_tool_calls=True)
    message = response.choices[0].message

    if message.tool_calls:
        print(f"Model requested {len(message.tool_calls)} tool call(s):")
        for i, tool_call in enumerate(message.tool_calls, 1):
            args = json.loads(tool_call.function.arguments)
            print(f"  {i}. {tool_call.function.name}: {args}")
    else:
        print(f"  Response: {message.content}")

    print("\n=== Example Complete ===")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Tests for deprecation warnings on .models property."""

import warnings
from unittest.mock import MagicMock, patch
import pytest

from esperanto.common_types import Model
from esperanto.providers.llm.base import LanguageModel
from esperanto.providers.embedding.base import EmbeddingModel
from esperanto.providers.reranker.base import RerankerModel
from esperanto.providers.stt.base import SpeechToTextModel
from esperanto.providers.tts.base import TextToSpeechModel


class TestLanguageModelDeprecation:
    """Test deprecation warnings for LanguageModel.models property."""

    def test_models_property_deprecated(self):
        """Test that .models property emits deprecation warning."""

        # Create a test implementation
        class TestLLM(LanguageModel):
            def _get_models(self):
                return [Model(id="test-model", owned_by="test")]

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="test")

        # Check that accessing .models raises a deprecation warning
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()
            assert "version 3.0" in str(w[0].message)
            assert "AIFactory.get_provider_models" in str(w[0].message)

    def test_models_property_still_works(self):
        """Test that .models property still returns correct data."""

        class TestLLM(LanguageModel):
            def _get_models(self):
                return [
                    Model(id="model1", owned_by="test"),
                    Model(id="model2", owned_by="test"),
                ]

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="test")

        # Suppress warnings for this test
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            models = model.models

            assert len(models) == 2
            assert models[0].id == "model1"
            assert models[1].id == "model2"


class TestEmbeddingModelDeprecation:
    """Test deprecation warnings for EmbeddingModel.models property."""

    def test_models_property_deprecated(self):
        """Test that .models property emits deprecation warning."""

        class TestEmbedding(EmbeddingModel):
            def _get_models(self):
                return [Model(id="test-embedding", owned_by="test")]

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def embed(self, texts, **kwargs):
                pass

            async def aembed(self, texts, **kwargs):
                pass

        model = TestEmbedding(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()


class TestRerankerModelDeprecation:
    """Test deprecation warnings for RerankerModel.models property."""

    def test_models_property_deprecated(self):
        """Test that .models property emits deprecation warning."""

        class TestReranker(RerankerModel):
            def _get_models(self):
                return [Model(id="test-reranker", owned_by="test")]

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def rerank(self, query, documents, **kwargs):
                pass

            async def arerank(self, query, documents, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestReranker(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()


class TestSTTModelDeprecation:
    """Test deprecation warnings for SpeechToTextModel.models property."""

    def test_models_property_deprecated(self):
        """Test that .models property emits deprecation warning."""

        class TestSTT(SpeechToTextModel):
            def _get_models(self):
                return [Model(id="test-stt", owned_by="test")]

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def transcribe(self, audio_file, **kwargs):
                pass

            async def atranscribe(self, audio_file, **kwargs):
                pass

        model = TestSTT(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()


class TestTTSModelDeprecation:
    """Test deprecation warnings for TextToSpeechModel.models property."""

    def test_models_property_deprecated(self):
        """Test that .models property emits deprecation warning."""

        class TestTTS(TextToSpeechModel):
            def _get_models(self):
                return [Model(id="test-tts", owned_by="test")]

            @property
            def provider(self):
                return "test"

            @property
            def available_voices(self):
                return {}

            def _get_default_model(self):
                return "test-model"

            def generate_speech(self, text, voice, **kwargs):
                pass

            async def agenerate_speech(self, text, voice, **kwargs):
                pass

        model = TestTTS(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()


class TestDeprecationMessage:
    """Test the specific content of deprecation messages."""

    def test_deprecation_message_includes_provider_name(self):
        """Test that deprecation message includes the provider name."""

        class TestLLM(LanguageModel):
            def _get_models(self):
                return []

            @property
            def provider(self):
                return "custom-provider"

            def _get_default_model(self):
                return "test-model"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            message = str(w[0].message)
            assert "custom-provider" in message
            assert "AIFactory.get_provider_models('custom-provider')" in message

    def test_deprecation_message_recommends_alternative(self):
        """Test that deprecation message recommends AIFactory method."""

        class TestLLM(LanguageModel):
            def _get_models(self):
                return []

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models

            message = str(w[0].message)
            assert "AIFactory.get_provider_models" in message
            assert "static" in message.lower() or "without creating provider instances" in message


class TestBackwardCompatibility:
    """Test that deprecated functionality still works for backward compatibility."""

    def test_old_code_continues_to_work(self):
        """Test that existing code using .models still works (with warnings)."""

        class TestLLM(LanguageModel):
            def _get_models(self):
                return [
                    Model(id="gpt-4", owned_by="openai"),
                    Model(id="gpt-3.5-turbo", owned_by="openai"),
                ]

            @property
            def provider(self):
                return "openai"

            def _get_default_model(self):
                return "gpt-4"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="gpt-4")

        # Old code pattern that should still work
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # This is how users currently access models
            available_models = model.models
            model_ids = [m.id for m in available_models]

            assert "gpt-4" in model_ids
            assert "gpt-3.5-turbo" in model_ids
            assert len(model_ids) == 2


class TestWarningStackLevel:
    """Test that warnings point to the correct code location."""

    def test_warning_points_to_user_code(self):
        """Test that warning stacklevel points to user's code, not library internals."""

        class TestLLM(LanguageModel):
            def _get_models(self):
                return []

            @property
            def provider(self):
                return "test"

            def _get_default_model(self):
                return "test-model"

            def chat_complete(self, messages, **kwargs):
                pass

            async def achat_complete(self, messages, **kwargs):
                pass

            def to_langchain(self):
                pass

        model = TestLLM(model_name="test")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            models = model.models  # This line should be in the warning

            assert len(w) == 1
            # The warning should have been raised from this test file, not from base.py
            assert "test_deprecation_warnings.py" in w[0].filename


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Universal transformers reranker provider with 4-strategy architecture."""

import asyncio
import os
import re
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

from esperanto.common_types import Model
from esperanto.common_types.reranker import RerankResponse, RerankResult
from .base import RerankerModel

if TYPE_CHECKING:
    import torch

# Optional transformers import with helpful error message
try:
    import torch
    from transformers import (
        AutoConfig,
        AutoModelForCausalLM,
        AutoModelForSequenceClassification,
        AutoTokenizer,
    )
    TRANSFORMERS_AVAILABLE = True
except ImportError:
    TRANSFORMERS_AVAILABLE = False
    torch = None
    AutoTokenizer = None
    AutoModelForCausalLM = None
    AutoModelForSequenceClassification = None
    AutoConfig = None

# Optional sentence_transformers import (part of transformers dependency)
try:
    from sentence_transformers import CrossEncoder
    SENTENCE_TRANSFORMERS_AVAILABLE = True
except ImportError:
    CrossEncoder = None
    SENTENCE_TRANSFORMERS_AVAILABLE = False

# Optional mxbai-rerank import
try:
    from mxbai_rerank import MxbaiRerankV2
    MXBAI_AVAILABLE = True
except ImportError:
    MxbaiRerankV2 = None
    MXBAI_AVAILABLE = False


# Define a no-op decorator when torch is not available
def no_grad_decorator(func):
    """Decorator that applies torch.no_grad() when available, otherwise returns function as-is."""
    if torch is not None and hasattr(torch, 'no_grad'):
        return torch.no_grad()(func)
    return func


@dataclass
class TransformersRerankerModel(RerankerModel):
    """Universal transformers-based reranker supporting multiple architectures.
    
    Supports 4 strategies:
    1. sentence_transformers - CrossEncoder models (cross-encoder/*, BAAI/bge-reranker-*, mixedbread v1)
    2. sequence_classification - AutoModelForSequenceClassification (jinaai/jina-reranker-*)
    3. causal_lm - AutoModelForCausalLM (Qwen/*-Reranker-*)
    4. mixedbread_v2 - mxbai-rerank library (mixedbread-ai/mxbai-rerank-*-v2)
    """
    
    device: Optional[str] = None
    cache_dir: Optional[str] = None
    trust_remote_code: bool = False

    # Model strategy patterns for auto-detection
    MODEL_STRATEGY_PATTERNS = {
        "cross-encoder/": "sentence_transformers",
        "BAAI/bge-reranker-": "sentence_transformers", 
        "mixedbread-ai/mxbai-rerank-base-v1": "sentence_transformers",
        "mixedbread-ai/mxbai-rerank-large-v1": "sentence_transformers", 
        "mixedbread-ai/mxbai-rerank-xsmall-v1": "sentence_transformers",
        "mixedbread-ai/mxbai-rerank-base-v2": "mixedbread_v2",
        "mixedbread-ai/mxbai-rerank-large-v2": "mixedbread_v2",
        "jinaai/jina-reranker-": "sequence_classification",
        "Qwen/": "causal_lm"
    }

    def __post_init__(self):
        """Initialize universal transformers reranker after dataclass initialization."""
        super().__post_init__()
        
        if not TRANSFORMERS_AVAILABLE:
            raise ImportError(
                "Transformers library not installed. Install with: pip install esperanto[transformers]"
            )
        
        # Set tokenizers parallelism to avoid fork warnings (especially in Jupyter)
        os.environ.setdefault("TOKENIZERS_PARALLELISM", "false")
        
        # Set cache directory if provided
        if self.cache_dir:
            os.environ["TRANSFORMERS_CACHE"] = self.cache_dir
        
        # Validate model name for security
        self._validate_model_name()
        
        # Auto-detect device
        self.device = self._detect_device()
        
        # Initialize model based on detected strategy
        self._load_model()

    def _validate_model_name(self):
        """Validate model name to prevent path traversal and other security issues."""
        model_name = self.get_model_name()
        
        # Check for path traversal attempts
        if ".." in model_name or "/" in model_name.replace("/", "", 1).replace("/", "", 1):
            # Allow at most one '/' for org/model format (e.g., "microsoft/model")
            parts = model_name.split("/")
            if len(parts) > 2:
                raise ValueError(f"Invalid model name format: {model_name}")
        
        # Check for invalid characters
        if not re.match(r"^[a-zA-Z0-9._/-]+$", model_name):
            raise ValueError(f"Model name contains invalid characters: {model_name}")
        
        # Check for suspicious patterns
        suspicious_patterns = ["\\", "..", "~", "$", "%", "@", "&", "|", ";", "`"]
        for pattern in suspicious_patterns:
            if pattern in model_name:
                raise ValueError(f"Model name contains suspicious pattern '{pattern}': {model_name}")

    def _detect_device(self) -> str:
        """Auto-detect the best available device."""
        if self.device:
            return self.device
            
        if torch.cuda.is_available():
            return "cuda"
        elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
            return "mps"
        else:
            return "cpu"

    def _get_model_strategy(self, model_name: str) -> str:
        """Automatically detect the correct strategy for the model."""
        # Check exact patterns first
        for pattern, strategy in self.MODEL_STRATEGY_PATTERNS.items():
            if model_name.startswith(pattern):
                return strategy
        
        # Fallback: inspect model config
        return self._detect_strategy_from_config(model_name)

    def _detect_strategy_from_config(self, model_name: str) -> str:
        """Detect strategy by inspecting model configuration."""
        try:
            config = AutoConfig.from_pretrained(model_name)
            
            # Map architectures to strategies
            if hasattr(config, 'architectures') and config.architectures:
                arch_str = str(config.architectures).lower()
                if "sequenceclassification" in arch_str:
                    return "sequence_classification"
                elif "causallm" in arch_str:
                    return "causal_lm"
                    
        except Exception:
            pass
        
        # If all detection fails
        supported_patterns = list(self.MODEL_STRATEGY_PATTERNS.keys())
        raise ValueError(
            f"Could not detect reranker strategy for model '{model_name}'. "
            f"Supported model patterns: {supported_patterns}"
        )

    def _load_model(self):
        """Load the appropriate model based on detected strategy."""
        model_name = self.get_model_name()
        strategy = self._get_model_strategy(model_name)
        
        if strategy == "sentence_transformers":
            self._load_sentence_transformers_model()
        elif strategy == "sequence_classification":
            self._load_sequence_classification_model()
        elif strategy == "causal_lm":
            self._load_causal_lm_model()
        elif strategy == "mixedbread_v2":
            self._load_mixedbread_v2_model()
        else:
            raise ValueError(f"Unknown reranker strategy: {strategy}")
        
        self.strategy = strategy

    def _load_sentence_transformers_model(self):
        """Load sentence_transformers CrossEncoder model."""
        if not SENTENCE_TRANSFORMERS_AVAILABLE:
            raise ImportError(
                "sentence-transformers library not available. Install with: pip install esperanto[transformers]"
            )
        
        try:
            self.model = CrossEncoder(self.get_model_name())
        except Exception as e:
            raise RuntimeError(f"Failed to load CrossEncoder model {self.get_model_name()}: {str(e)}")

    def _load_sequence_classification_model(self):
        """Load AutoModelForSequenceClassification model."""
        try:
            model_name = self.get_model_name()
            
            self.model = AutoModelForSequenceClassification.from_pretrained(
                model_name,
                torch_dtype="auto",
                trust_remote_code=self.trust_remote_code,
                cache_dir=self.cache_dir
            )
            self.tokenizer = AutoTokenizer.from_pretrained(
                model_name,
                cache_dir=self.cache_dir
            )
            
            self.model.to(self.device)
            self.model.eval()
            
        except Exception as e:
            raise RuntimeError(f"Failed to load sequence classification model {self.get_model_name()}: {str(e)}")

    def _load_causal_lm_model(self):
        """Load AutoModelForCausalLM model (Qwen style)."""
        try:
            model_name = self.get_model_name()
            
            # Load tokenizer with left padding (required for Qwen reranker)
            self.tokenizer = AutoTokenizer.from_pretrained(
                model_name,
                cache_dir=self.cache_dir,
                padding_side='left'
            )
            
            # Load model as causal LM
            self.model = AutoModelForCausalLM.from_pretrained(
                model_name,
                cache_dir=self.cache_dir,
                torch_dtype=torch.float16 if self.device in ["cuda", "mps"] else torch.float32
            )
            
            # Move model to device
            self.model.to(self.device)
            self.model.eval()
            
            # Setup Qwen-specific configuration
            self._setup_qwen_reranker()
            
        except Exception as e:
            raise RuntimeError(f"Failed to load causal LM model {self.get_model_name()}: {str(e)}")

    def _setup_qwen_reranker(self):
        """Setup Qwen-specific reranker configuration."""
        # Get token IDs for "yes" and "no" 
        self.token_false_id = self.tokenizer.convert_tokens_to_ids("no")
        self.token_true_id = self.tokenizer.convert_tokens_to_ids("yes")
        
        # Define prompt templates
        self.prefix = "<|im_start|>system\nJudge whether the Document meets the requirements based on the Query and the Instruct provided. Note that the answer can only be \"yes\" or \"no\".<|im_end|>\n<|im_start|>user\n"
        self.suffix = "<|im_end|>\n<|im_start|>assistant\n<think>\n\n</think>\n\n"
        
        # Encode prefix and suffix tokens
        self.prefix_tokens = self.tokenizer.encode(self.prefix, add_special_tokens=False)
        self.suffix_tokens = self.tokenizer.encode(self.suffix, add_special_tokens=False)
        
        # Set max length for the model
        self.max_length = 8192

    def _load_mixedbread_v2_model(self):
        """Load Mixedbread v2 models with optional dependency."""
        if not MXBAI_AVAILABLE:
            raise ImportError(
                "mxbai-rerank library required for Mixedbread v2 models. "
                "Install with: pip install mxbai-rerank"
            )
        
        try:
            self.model = MxbaiRerankV2(self.get_model_name())
        except Exception as e:
            raise RuntimeError(f"Failed to load Mixedbread v2 model {self.get_model_name()}: {str(e)}")

    def _score_all_pairs(self, query: str, documents: List[str]) -> List[float]:
        """Universal scoring method that dispatches to appropriate strategy."""
        if self.strategy == "sentence_transformers":
            return self._rerank_sentence_transformers(query, documents)
        elif self.strategy == "sequence_classification":
            return self._rerank_sequence_classification(query, documents)
        elif self.strategy == "causal_lm":
            return self._rerank_causal_lm(query, documents)
        elif self.strategy == "mixedbread_v2":
            return self._rerank_mixedbread_v2(query, documents)
        else:
            raise ValueError(f"Unknown reranker strategy: {self.strategy}")

    def _rerank_sentence_transformers(self, query: str, documents: List[str]) -> List[float]:
        """Rerank using sentence_transformers CrossEncoder."""
        try:
            # Use CrossEncoder.rank method
            ranks = self.model.rank(query, documents)
            
            # Reorder scores to match original document order
            ordered_scores = [0.0] * len(documents)
            for rank in ranks:
                ordered_scores[rank['corpus_id']] = rank['score']
            
            return ordered_scores
            
        except Exception as e:
            import logging
            logging.warning(f"CrossEncoder reranker error: {str(e)}")
            return [0.0] * len(documents)

    def _rerank_sequence_classification(self, query: str, documents: List[str]) -> List[float]:
        """Rerank using AutoModelForSequenceClassification."""
        try:
            # Format query-document pairs 
            sentence_pairs = [[query, doc] for doc in documents]
            
            # Compute scores using model's compute_score method
            scores = self.model.compute_score(sentence_pairs, max_length=1024)
            
            return scores.tolist() if hasattr(scores, 'tolist') else list(scores)
            
        except Exception as e:
            import logging
            logging.warning(f"Sequence classification reranker error: {str(e)}")
            return [0.0] * len(documents)

    def _rerank_causal_lm(self, query: str, documents: List[str]) -> List[float]:
        """Rerank using AutoModelForCausalLM (Qwen style)."""
        try:
            # Format all query-document pairs using Qwen instruction format
            instruction = 'Given a web search query, retrieve relevant passages that answer the query'
            pairs = [
                self._format_instruction(instruction, query, doc) 
                for doc in documents
            ]
            
            # Process inputs and compute scores
            inputs = self._process_inputs(pairs)
            scores = self._compute_logits(inputs)
            
            return scores
            
        except Exception as e:
            import logging
            logging.warning(f"Causal LM reranker error: {str(e)}")
            return [0.0] * len(documents)

    def _rerank_mixedbread_v2(self, query: str, documents: List[str]) -> List[float]:
        """Rerank using mxbai-rerank library."""
        try:
            # Use mxbai library rank method
            results = self.model.rank(query, documents, return_documents=True, top_k=len(documents))
            
            # Extract scores and reorder to match original document order
            scores = [0.0] * len(documents)
            for i, result in enumerate(results):
                # mxbai returns RankResult objects with attributes, not dicts
                original_index = getattr(result, 'corpus_id', None)
                score = getattr(result, 'score', 0.0)
                
                if original_index is not None and 0 <= original_index < len(scores):
                    scores[original_index] = score
                else:
                    # Handle missing or invalid corpus_id
                    import logging
                    logging.warning(f"Invalid or missing corpus_id for result {i}: {original_index}")
            
            return scores
            
        except Exception as e:
            import logging
            logging.warning(f"Mixedbread v2 reranker error: {str(e)}")
            return [0.0] * len(documents)

    def _format_instruction(self, instruction: Optional[str], query: str, doc: str) -> str:
        """Format the instruction for Qwen reranker."""
        if instruction is None:
            instruction = 'Given a web search query, retrieve relevant passages that answer the query'
        return f"<Instruct>: {instruction}\n<Query>: {query}\n<Document>: {doc}"

    def _process_inputs(self, pairs: List[str]) -> Dict[str, "torch.Tensor"]:
        """Process input pairs for Qwen reranker."""
        # Build full text strings with prefix and suffix for direct tokenization
        # This avoids the warning about using encode() followed by pad()
        prefix_text = self.tokenizer.decode(self.prefix_tokens, skip_special_tokens=False)
        suffix_text = self.tokenizer.decode(self.suffix_tokens, skip_special_tokens=False)
        
        full_texts = [prefix_text + pair + suffix_text for pair in pairs]
        
        # Use the faster __call__ method as recommended by the warning
        inputs = self.tokenizer(
            full_texts,
            padding=True,
            truncation=True,
            max_length=self.max_length,
            return_tensors="pt",
            return_attention_mask=True
        )
        
        # Move to device
        for key in inputs:
            inputs[key] = inputs[key].to(self.device)
        
        return inputs

    @no_grad_decorator
    def _compute_logits(self, inputs: Dict[str, "torch.Tensor"]) -> List[float]:
        """Compute logits and extract relevance scores for Qwen."""
        batch_scores = self.model(**inputs).logits[:, -1, :]
        true_vector = batch_scores[:, self.token_true_id]
        false_vector = batch_scores[:, self.token_false_id]
        batch_scores = torch.stack([false_vector, true_vector], dim=1)
        batch_scores = torch.nn.functional.log_softmax(batch_scores, dim=1)
        scores = batch_scores[:, 1].exp().tolist()
        return scores

    def rerank(
        self, 
        query: str, 
        documents: List[str], 
        top_k: Optional[int] = None, 
        **kwargs
    ) -> RerankResponse:
        """Rerank documents using universal transformers model.
        
        Args:
            query: The search query to rank documents against.
            documents: List of documents to rerank.
            top_k: Maximum number of results to return.
            **kwargs: Additional arguments.
            
        Returns:
            RerankResponse with ranked results.
        """
        # Validate inputs
        query, documents, top_k = self._validate_inputs(query, documents, top_k)
        
        # Score all query-document pairs using appropriate strategy
        raw_scores = self._score_all_pairs(query, documents)
        
        # Normalize scores using base class method
        normalized_scores = self._normalize_scores(raw_scores)
        
        # Create results with original indices
        results = []
        for i, (document, score) in enumerate(zip(documents, normalized_scores)):
            results.append(RerankResult(
                index=i,
                document=document,
                relevance_score=score
            ))
        
        # Sort by relevance score (highest first)
        results.sort(key=lambda x: x.relevance_score, reverse=True)
        
        # Apply top_k limit
        if top_k < len(results):
            results = results[:top_k]
        
        return RerankResponse(
            results=results,
            model=self.get_model_name(),
            usage=None  # Local models don't have usage stats
        )

    async def arerank(
        self, 
        query: str, 
        documents: List[str], 
        top_k: Optional[int] = None, 
        **kwargs
    ) -> RerankResponse:
        """Async rerank documents using universal transformers model.
        
        Args:
            query: The search query to rank documents against.
            documents: List of documents to rerank.
            top_k: Maximum number of results to return.
            **kwargs: Additional arguments.
            
        Returns:
            RerankResponse with ranked results.
        """
        # Run the sync rerank method in a thread pool
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as executor:
            result = await loop.run_in_executor(
                executor, 
                self.rerank, 
                query, 
                documents, 
                top_k
            )
        return result

    def to_langchain(self):
        """Convert to LangChain-compatible reranker."""
        try:
            from langchain_core.documents import Document
            from langchain_core.callbacks.manager import Callbacks
        except ImportError:
            raise ImportError(
                "LangChain not installed. Install with: pip install langchain"
            )
        
        class LangChainUniversalReranker:
            def __init__(self, transformers_reranker):
                self.transformers_reranker = transformers_reranker
            
            def compress_documents(
                self, 
                documents: List[Document], 
                query: str, 
                callbacks: Optional[Callbacks] = None
            ) -> List[Document]:
                """Compress documents using universal transformers reranker."""
                # Note: callbacks parameter is part of LangChain interface but not used in local processing
                _ = callbacks  # Acknowledge parameter to avoid warnings
                
                # Extract text content from documents
                texts = [doc.page_content for doc in documents]
                
                # Rerank using transformers
                rerank_response = self.transformers_reranker.rerank(query, texts)
                
                # Convert back to LangChain documents
                reranked_docs = []
                for result in rerank_response.results:
                    if result.index < len(documents):
                        original_doc = documents[result.index]
                        # Add relevance score to metadata
                        new_metadata = original_doc.metadata.copy()
                        new_metadata["relevance_score"] = result.relevance_score
                        
                        reranked_docs.append(Document(
                            page_content=original_doc.page_content,
                            metadata=new_metadata
                        ))
                
                return reranked_docs
        
        return LangChainUniversalReranker(self)

    def _get_default_model(self) -> str:
        """Get default transformers model."""
        return "Qwen/Qwen3-Reranker-4B"

    @property
    def provider(self) -> str:
        """Provider name."""
        return "transformers"

    def _get_models(self) -> List[Model]:
        """Available universal transformers reranker models."""
        return [
            # Qwen models (causal_lm strategy)
            Model(
                id="Qwen/Qwen3-Reranker-4B",
                owned_by="Qwen",
                context_window=8192),
            Model(
                id="Qwen/Qwen3-Reranker-0.6B",
                owned_by="Qwen",
                context_window=8192),
            # CrossEncoder models (sentence_transformers strategy)
            Model(
                id="cross-encoder/ms-marco-MiniLM-L-6-v2",
                owned_by="microsoft",
                context_window=512),
            Model(
                id="cross-encoder/ms-marco-electra-base",
                owned_by="microsoft",
                context_window=512),
            Model(
                id="BAAI/bge-reranker-base",
                owned_by="BAAI",
                context_window=512),
            Model(
                id="BAAI/bge-reranker-large",
                owned_by="BAAI",
                context_window=512),
            # Mixedbread v1 models (sentence_transformers strategy)
            Model(
                id="mixedbread-ai/mxbai-rerank-xsmall-v1",
                owned_by="mixedbread-ai",
                context_window=512),
            Model(
                id="mixedbread-ai/mxbai-rerank-base-v1",
                owned_by="mixedbread-ai",
                context_window=512),
            Model(
                id="mixedbread-ai/mxbai-rerank-large-v1",
                owned_by="mixedbread-ai",
                context_window=512),
            # Jina models (sequence_classification strategy)
            Model(
                id="jinaai/jina-reranker-v2-base-multilingual",
                owned_by="jinaai",
                context_window=1024),
            # Mixedbread v2 models (mixedbread_v2 strategy)
            Model(
                id="mixedbread-ai/mxbai-rerank-base-v2",
                owned_by="mixedbread-ai",
                context_window=512),
            Model(
                id="mixedbread-ai/mxbai-rerank-large-v2",
                owned_by="mixedbread-ai",
                context_window=512),
        ]